/*
 * Copyright (C) 1998, 1999, 2000, 2001, 2002
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 * 
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.  
 */ 
 
/*----------------------------------------------------------------
 * project ....: LTI Digital Image/Signal Processing Library
 * file .......: ltikNNClassifier.h
 * authors ....: Frederik Lange
 * organization: LTI, RWTH Aachen
 * creation ...: 15.06.2003
 * revisions ..: $Id: ltiKNNClassifier.h,v 1.3 2003/11/19 06:55:59 alvarado Exp $
 */

#ifndef _LTI_KNN_CLASSIFIER_H_
#define _LTI_KNN_CLASSIFIER_H_

#include "ltiSupervisedInstanceClassifier.h"
#include "ltiObjectFactory.h"
#include "ltiKernelFunctor.h"
#include "ltiClassName.h"
#include "ltiBoundsFunctor.h"
#include "ltiKdTree.h"

#include <map>
#include <vector>

namespace lti {
  /**
   * Implements a k nearest neighbors search based classifier.
   *
   * The simplest case of a k-nearest neighbor classifier is for k=1, also
   * known as a nearest neighbor classifier, which assings as winner class for
   * a test point \a x the one belonging to the nearest sample point.
   *
   * For k>1, a k nearest neighbor classifier assigns to a point \a x the class
   * most represented in the k nearest neighbors.  In the simplest case, each
   * of the k nearest sample points vote with the same weight for their
   * respective classes.  In more sophisticated cases, each point will vote
   * with a weight depending on the total number of sample points of its class
   * and/or the ratio between the distance of the test point to the winner
   * sample and the distance of the test point to the first sample point
   * belonging to another class.
   *
   * This functor allows at this point only the use of Euclidean distance.
   *
   * This classifier uses a kd-Tree to perform the search in an efficient
   * manner, but shows therefore also the drawbacks of a normal kd-Tree: it
   * is not suitable for higher dimensional spaces.
   *
   * Example:
   *
   * \code
   *
   * #include <iostream>
   * #include <fstream>
   *
   * #include "ltiKNNClassifier.h"
   * #include "ltiLispStreamHandler.h"
   *
   * // ...
   *
   * double inData[] = {-1,-1,
   *                    -1, 0,
   *                    -1,+1,
   *                    +0,+1,
   *                    +1,+1,
   *                    +1,+0,
   *                    +1,-1,
   *                    +0,-1,
   *                    +0,+0};
   *
   * lti::dmatrix inputs(9,2,inData);     // training vectors
   *
   * int idsData[] = {1,0,2,1,0,2,1,0,1}; // and the respective ids
   * lti::ivector ids(9,idsData);
   *
   * lti::kNNClassifier knn;  // our nearest neighbor classifier
   *
   * lti::kNNClassifier::parameters param;
   * param.useReliabilityMeasure = false;
   * knn.setParameters(param);
   *
   * // we want to see some info while training
   * streamProgressInfo prog(std::cout);
   * knn.setProgressObject(prog);
   *
   * // train the network
   * knn.train(inputs,ids);
   *
   * // let us save our network for future use
   * // in the file called mlp.dat
   * std::ofstream out("knn.dat");
   * lti::lispStreamHandler lsh(out);
   *
   * // save the network
   * knn.write(lsh);
   * // close the file
   * out.close();
   *
   * // show some results with the same training set:
   *
   * lti::kNNClassifier::outputVector outv; // here we will get some
   *                                        // classification results
   * std::cout << std::endl << "Results: " << std::endl;
   *
   * int id;
   * dvector sample(2,0.0);
   * // generate some points and check which would be the winner class
   * for (sample.at(0) = -1; sample.at(0) <= 1; sample.at(0)+=0.25) {
   *   for (sample.at(1) = -1; sample.at(1) <= 1; sample.at(1)+=0.25) {
   * 
   *     knn.classify(sample,outv);
   *     std::cout << "Input " << sample << " \tOutput: ";
   *     outv.getId(outv.getWinnerUnit(),id);
   *     std::cout << id;
   *     std::cout << std::endl;
   *   }
   * }
   *
   * \endcode
   */
  class kNNClassifier : public supervisedInstanceClassifier {
  public:

    /**
     * the parameters for the class kNNClassifier
     */
    class parameters : public supervisedInstanceClassifier::parameters {
    public:

      /**
       * default constructor
       */
      parameters();

      /**
       * copy constructor
       * @param other the parameters object to be copied
       */
      parameters(const parameters& other);

      /**
       * destructor
       */
      virtual ~parameters();

      /**
       * returns name of this type.
       */
      const char* getTypeName() const;

      /**
       * copy the contents of a parameters object
       * @param other the parameters object to be copied
       * @return a reference to this parameters object
       */
      parameters& copy(const parameters& other);

      /**
       * copy the contents of a parameters object
       * @param other the parameters object to be copied
       * @return a reference to this parameters object
       */
      parameters& operator=(const parameters& other);


      /**
       * returns a pointer to a clone of the parameters
       */
      virtual classifier::parameters* clone() const;

      /**
       * write the parameters in the given ioHandler
       * @param handler the ioHandler to be used
       * @param complete if true (the default) the enclosing begin/end will
       *        be also written, otherwise only the data block will be written.
       * @return true if write was successful
       */
      virtual bool write(ioHandler& handler,const bool& complete=true) const;

      /**
       * read the parameters from the given ioHandler
       * @param handler the ioHandler to be used
       * @param complete if true (the default) the enclosing begin/end will
       *        be also written, otherwise only the data block will be written.
       * @return true if write was successful
       */
      virtual bool read(ioHandler& handler,const bool& complete=true);

#     ifdef _LTI_MSC_VER
      /**
       * this function is required by MSVC only, as a workaround for a
       * very awful bug, which exists since MSVC V.4.0, and still by
       * V.6.0 with all bugfixes (so called "service packs") remains
       * there...  This method is also public due to another bug, so please
       * NEVER EVER call this method directly: use read() instead
       */
      bool readMS(ioHandler& handler,const bool& complete=true);

      /**
       * this function is required by MSVC only, as a workaround for a
       * very awful bug, which exists since MSVC V.4.0, and still by
       * V.6.0 with all bugfixes (so called "service packs") remains
       * there...  This method is also public due to another bug, so please
       * NEVER EVER call this method directly: use write() instead
       */
      bool writeMS(ioHandler& handler,const bool& complete=true) const;
#     endif

      // ------------------------------------------------
      // the parameters
      // ------------------------------------------------

      /**
       * How many neirest neighbors should be determined per classification.
       *
       * Default value: 1 (i.e. nearest neighbor classifier)
       */
      int kNN;

      /**
       * Normalize data to equal number of data samples.
       * 
       * The number of samples per class may differ.  Therefore samples can be
       * weighted relative to the number of samples per class.
       *
       * Default value is true
       */
      bool normalizeData;

      /**
       * @name Reliability
       */
      //@{
      /**
       * Use the reliability measure suggested by Lowe. 
       *
       * Lowe suggested in his paper: "Distinctive Image Features from
       * Scale Invariant Keypoints", June, 2003 the use of a
       * reliability measure for a classification.  It is defined as
       * the ratio between the distance from the analyzed point \a p
       * to the closest sample point \a w, and the distance from the
       * same point \a p to the closest point that belongs to a class
       * different to the one of the point \a w.
       *
       * You usually use this mode with kNN=1.
       * 
       * Default value: false
       */
      bool useReliabilityMeasure;

      /**
       * If useReliabilityMeasure is true, then the weight of a point
       * can be determined using the ratio of two distances, but there
       * are many possiblities to consider this ration.  Several simple
       * modes are provided here.  Let \e d1 be the distance to the
       * winner sample point and \e d2 the distance to the closest point
       * belonging to a different class than the winner.
       */
      enum eReliabilityMode {
        Linear,      /**< The weight is determined as 
                          min(1.0,((d2/d1)-1)/(threshold-1))).  The threshold
                          must be strict greater than 1.0 */
        Exponential, /**< The weight is determined as
                          1.0 - exp(-(d2/d1 - 1)/threshold) */
      };

      /**
       * Reliability mode used.
       *
       * For possible values see eReliabilityMode.
       *
       * Default value: Linear
       */
      eReliabilityMode reliabilityMode;

      /**
       * Threshold value used for the reliability function
       *
       * Default value: 10, i.e. distances ratios greater than d2/d1 should be
       *                    consider with the same weight.
       */
      double reliabilityThreshold;

      /**
       * Maximal number of neighbors considered while detecting the
       * second point belonging to another class than the winner.  If
       * no point was found within this number of points, a "perfectly"
       * reliable point will be assumed.
       *
       * Default value: 20
       */
      int maxUnreliableNeighborhood;

      //@}

      /**
       * @name Nearest neighbor search options
       *
       * The search is optimized using a kd-Tree.  The parameters in
       * this group control how the tree is organized or how to search
       * for the data therein.
       */
      //@{
      /**
       * Best Bin First.
       *
       * If this parameter is set to true, the Best Bin First (BBF) algorithm
       * of Lowe et. al will be applied.  It is an approximative algorithm
       * appropriate for spaces of relative high dimensionality (100 or so)
       * in which some improbable bins are discarded in the search.
       *
       * Note that if you activate this, the result is possible optimal, but
       * not optimal.
       *
       * Default value: false
       */
      bool bestBinFirst;

      /**
       * Maximum visit number per leaf node allowed.
       *
       * This is only required if you specify a best-bin-first (BBF)
       * search.  It is the maximal number of visits allowed for leaf
       * nodes. (in the original paper is called Emax).  
       *
       * Usually this value depends on many factors.  You can set it as a
       * percentage of the expected number of leaf nodes (about number of
       * points/bucketSize).
       *
       * Default value: 100, but you should set it in order for BBF to work
       *                     appropriatelly.
       */
      int eMax;

      /**
       * Bucket size.
       *
       * Each node of the kd-Tree can contain several points.  The search
       * within a node is made with linear search, i.e. the brute force method
       * computing the distances to each single point.  This parameter 
       * indicates the number of points that will be stored in the node.
       * The name "bucket" comes from the original kd-Tree paper.
       *
       * Default value: 5
       */
      int bucketSize;
      //@}
    };

    /**
     * default constructor
     */
    kNNClassifier();

    /**
     * copy constructor
     * @param other the object to be copied
     */
    kNNClassifier(const kNNClassifier& other);

    /**
     * destructor
     */
    virtual ~kNNClassifier();

    /**
     * returns the name of this type ("kNNClassifier")
     */
    virtual const char* getTypeName() const;

    /**
     * copy data of "other" clustering.
     * @param other the clustering to be copied
     * @return a reference to this clustering object
     */
    kNNClassifier& copy(const kNNClassifier& other);

    /**
     * alias for copy member
     * @param other the clustering to be copied
     * @return a reference to this clustering object
     */
    kNNClassifier& operator=(const kNNClassifier& other);

    /**
     * returns a pointer to a clone of this clustering.
     */
    virtual classifier* clone() const;

    /**
     * returns used parameters
     */
    const kNNClassifier::parameters& getParameters() const;

    /**
     * Supervised training.  The vectors in the <code>input</code> matrix
     * are arranged row-wise, i.e. each row contains one data vector.
     * The <code>ids</code> vector contains the class label for each row.
     *
     * This is an alternative method to trainObject().  You cannot add further 
     * objects after you have called train(), nor can you call train()
     * after calling trainObject(), since all data provided with trainObject()
     * would be removed.  In other words, you must decide if you want to
     * supply all objects separately or if you want to give them all
     * simultaneously, but you cannot combine both methods.
     *
     * @param input the matrix with the input vectors (each row is a training
     *              vector)
     * @param ids vector of class ids for each input point
     *
     * @return true if successful, false otherwise. (if false you can check
     *              the error message with getStatusString())
     */
    virtual bool train(const dmatrix& input,
                       const ivector& ids);

    /**
     * Adds an object to this classifier. The id is given automatically
     * and returned in the parameter.
     *
     * After you have trained several objects, you must call the build() 
     * method to finish the training process.  If you don't do it, the 
     * classifier will ignore everything you have provided.
     *
     * This is an alternative method to train().  You cannot add further 
     * objects after you have called train, nor can you call train()
     * after calling this method, since all data provided with trainObject
     * would be removed.  In other words, you must decide if you want to
     * supply all objects separately or if you want to give them all
     * simultaneously, but you cannot combine both methods.
     */
    virtual bool trainObject(const dmatrix& input, int& id);

    /**
     * Classification.
     *
     * Classifies the feature and returns the outputVector %object with
     * the classification result.
     *
     * <b>NOTE:</b> This method is NOT really const. Although the main
     * members of the kNNClassifier are not changed some state variables used
     * for efficiency are. Thus, it is not save to use the same instance of
     * the kNNClassifier in two different threads.
     *
     * @param feature pattern to be classified
     * @param result of the classifications as a classifier::outputVector
     * @return true if the classification has been successful
     */
    virtual bool classify(const dvector& feature, outputVector& result) const;
    
    /**
     * Classification.
     *
     * Classifies all features (the rows of the matrix) and returns
     * the outputVector %object with the classification result.
     *
     * The classification will be the accumulation of the voting for all
     * given points, assuming that they all belong to the same class.
     *
     * <b>NOTE:</b> This method is NOT really const. Although the main members
     * of the kNNClassifier are not changed some state variables used for
     * efficiency are. Thus, it is not save to use the same instance of the
     * kNNClassifier in two different threads.
     *
     * @param features patterns to be classified each row is one feature
     * @param result of the classifications as a classifier::outputVector
     * @return true if the classification has been successful
     */
    virtual bool classify(const dmatrix& features, outputVector& result) const;

    /**
     * write the classifier in the given ioHandler
     * @param handler the ioHandler to be used
     * @param complete if true (the default) the enclosing begin/end will
     *        be also written, otherwise only the data block will be written.
     * @return true if write was successful
     */
    virtual bool write(ioHandler& handler,const bool& complete=true) const;
    
    /**
     * read the classifier from the given ioHandler
     * @param handler the ioHandler to be used
     * @param complete if true (the default) the enclosing begin/end will
     *        be also written, otherwise only the data block will be written.
     * @return true if write was successful
     */
    virtual bool read(ioHandler& handler,const bool& complete=true);
    
    /**
     * You must call this method if you have added single data sets, in order
     * to complete the training process.
     *
     * @see trainObject(),reset()
     */
    void build();
    
    /**
     * Resets all values and deletes the content.
     *
     * If you want to forget the sample points and start giving new points
     * with trainObject, you need to call this method first
     */
    void clear();
  
  protected:

    /**
     * @name Reliability weighting functions
     */
    //@{
    static double linear(const double& ratio,const double& threshold);
    static double exponential(const double& ratio,const double& threshold);
    //@}

    /**
     *  mapping of the ids
     */    
    void buildIdMaps(const ivector& ids);
    
    /**
     * Define the output template
     */
    void defineOutputTemplate();

    /**
     * Type for maps mapping ids from internal to external and viceversa
     */
    typedef std::map<int,int> idMap_type;

    /**
     * Map from external id to internal id, used for training
     */
    idMap_type idMap;

    /**
     * Map from internal to external id, used for training
     */
    idMap_type rIdMap;

    /**
     * Number of classes currently in the classifier
     */
    int nClasses;

    /**
     * Exact kd-Tree type used for the database
     */
    typedef kdTree<dvector,int> treeType;
    

    /**
     * The database with accelerated nearest neighbor search.
     *
     * The kdTree uses as n-dimensional points dvectors and as data requires
     * only the id of each vector.
     */
    treeType databaseTree;

    /**
     * Optionally a class weight can be applied, as a-priori value.
     *
     * The std::vector is used due to the push_back interface.
     */
    std::vector<double> classWeight;

    /**
     * Minimum number of points per class.
     *
     * This attribute is valid only after the complete training process
     */
    int minPointsPerClass;

    /**
     * Maximum number of points per class
     *
     * This attribute is valid only after the complete training process
     */
    int maxPointsPerClass;
  };

}

#endif
