/*
 * Copyright (C) 1998, 1999, 2000, 2001
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 *
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.
 */


/* -------------------------------------------------------------------
 * project ....: LTI-Lib: Image Processing and Computer Vision Library
 * file .......: ltiCannyEdges.cpp
 * authors ....: Christian Harte
 * organization: LTI, RWTH Aachen
 * creation ...: 19.7.2002
 * revisions ..: $Id: ltiCannyEdges.cpp,v 1.25 2003/11/14 14:15:19 alvarado Exp $
 */

#include "ltiCannyEdges.h"
#include "ltiGaussKernels.h"
#include "ltiGradientKernels.h"
#include "ltiConvolution.h"
#include "ltiSplitImageToRGB.h"
#include "ltiMaximumFilter.h"
#include "ltiConstants.h"
#include "ltiColorContrastGradient.h"

#include <limits>

#undef _LTI_DEBUG
//#define _LTI_DEBUG 2
#include "ltiDebug.h"

#ifdef _LTI_DEBUG
#include "ltiViewer.h"
#endif

namespace lti {
  // --------------------------------------------------
  // cannyEdges::parameters
  // --------------------------------------------------

  // default constructor
  cannyEdges::parameters::parameters()
    : edgeDetector::parameters() {

    variance = float(1.0f);
    kernelSize = 7;
    thresholdMin = float(0.5f);
    thresholdMax = float(0.04f);
    optimalGradientKernelSize=0;
  }

  // copy constructor
  cannyEdges::parameters::parameters(const parameters& other)
    : edgeDetector::parameters()  {
    copy(other);
  }

  // destructor
  cannyEdges::parameters::~parameters() {
  }

  // get type name
  const char* cannyEdges::parameters::getTypeName() const {
    return "cannyEdges::parameters";
  }

  // copy member

  cannyEdges::parameters&
    cannyEdges::parameters::copy(const parameters& other) {
# ifndef _LTI_MSC_VER
    // MS Visual C++ 6 is not able to compile this...
    edgeDetector::parameters::copy(other);
# else
    // ...so we have to use this workaround.
    // Conditional on that, copy may not be virtual.
    edgeDetector::parameters& (edgeDetector::parameters::* p_copy)
      (const edgeDetector::parameters&) =
      edgeDetector::parameters::copy;
    (this->*p_copy)(other);
# endif

    variance = other.variance;
    kernelSize = other.kernelSize;
    thresholdMin = other.thresholdMin;
    thresholdMax = other.thresholdMax;
    optimalGradientKernelSize = other.optimalGradientKernelSize;
    return *this;
  }

  // alias for copy member
  cannyEdges::parameters&
    cannyEdges::parameters::operator=(const parameters& other) {
    return copy(other);
  }

  // clone member
  functor::parameters* cannyEdges::parameters::clone() const {
    return new parameters(*this);
  }

  /*
   * write the parameters in the given ioHandler
   * @param handler the ioHandler to be used
   * @param complete if true (the default) the enclosing begin/end will
   *        be also written, otherwise only the data block will be written.
   * @return true if write was succeful
   */
# ifndef _LTI_MSC_VER
  bool cannyEdges::parameters::write(ioHandler& handler,
                                         const bool& complete) const
# else
  bool cannyEdges::parameters::writeMS(ioHandler& handler,
                                           const bool& complete) const
# endif
  {
    bool b = true;
    if (complete) {
      b = handler.writeBegin();
    }

    if (b) {

      lti::write(handler,"variance",variance);
      lti::write(handler,"kernelSize",kernelSize);
      lti::write(handler,"thresholdMin",thresholdMin);
      lti::write(handler,"thresholdMax",thresholdMax);
      lti::write(handler,"optimalGradientKernelSize",
                 optimalGradientKernelSize);
    }

# ifndef _LTI_MSC_VER
    // This is the standard C++ code, which MS Visual C++ 6 is not able to
    // compile...
    b = b && edgeDetector::parameters::write(handler,false);
# else
    bool (edgeDetector::parameters::* p_writeMS)(ioHandler&,const bool&) const =
      edgeDetector::parameters::writeMS;
    b = b && (this->*p_writeMS)(handler,false);
# endif

    if (complete) {
      b = b && handler.writeEnd();
    }

    return b;
  }

# ifdef _LTI_MSC_VER
  bool cannyEdges::parameters::write(ioHandler& handler,
                                         const bool& complete) const {
    // ...we need this workaround to cope with another really awful MSVC bug.
    return writeMS(handler,complete);
  }
# endif

  /*
   * read the parameters from the given ioHandler
   * @param handler the ioHandler to be used
   * @param complete if true (the default) the enclosing begin/end will
   *        be also read, otherwise only the data block will be read.
   * @return true if write was succeful
   */
# ifndef _LTI_MSC_VER
  bool cannyEdges::parameters::read(ioHandler& handler,
                                        const bool& complete)
# else
  bool cannyEdges::parameters::readMS(ioHandler& handler,
                                          const bool& complete)
# endif
  {
    bool b = true;
    if (complete) {
      b = handler.readBegin();
    }

    if (b) {
      lti::read(handler,"variance",variance);
      lti::read(handler,"kernelSize",kernelSize);
      lti::read(handler,"thresholdMin",thresholdMin);
      lti::read(handler,"thresholdMax",thresholdMax);
      lti::read(handler,"optimalGradientKernelSize",
                optimalGradientKernelSize);
      
    }

# ifndef _LTI_MSC_VER
    // This is the standard C++ code, which MS Visual C++ 6 is not able to
    // compile...
    b = b && edgeDetector::parameters::read(handler,false);
# else
    bool (edgeDetector::parameters::* p_readMS)(ioHandler&,const bool&) =
      edgeDetector::parameters::readMS;
    b = b && (this->*p_readMS)(handler,false);
# endif

    if (complete) {
      b = b && handler.readEnd();
    }

    return b;
  }

# ifdef _LTI_MSC_VER
  bool cannyEdges::parameters::read(ioHandler& handler,
                                        const bool& complete) {
    // ...we need this workaround to cope with another really awful MSVC
    // bug.
    return readMS(handler,complete);
  }
# endif

  // --------------------------------------------------
  // cannyEdges
  // --------------------------------------------------

  /*
   * convert a pair of values in cartesic system into a pair in
   * polar system
   */
  inline void cannyEdges::cartToPolar(const int dx,const int dy,
                                      float& mag,float& angle) const {
    static const float factor = 1.0f/255.0f;
    mag = factor*sqrt(float(dx*dx+dy*dy));
    angle = atan2(dy,dx);
  }

  // default constructor
  cannyEdges::cannyEdges()
    : edgeDetector(){

    // create an instance of the parameters with the default values
    parameters defaultParameters;
    // set the default parameters
    setParameters(defaultParameters);

  }

  // default constructor
  cannyEdges::cannyEdges(const parameters& thePars)
    : edgeDetector() {

    // set the default parameters
    setParameters(thePars);
  }


  // copy constructor
  cannyEdges::cannyEdges(const cannyEdges& other)
    : edgeDetector() {

    copy(other);
  }

  // destructor
  cannyEdges::~cannyEdges() {
  }

  // returns the name of this type
  const char* cannyEdges::getTypeName() const {
    return "cannyEdges";
  }

  // which apply is faster
  functor::eFasterApply cannyEdges::theFaster() const {
    return OnCopy;
  }

  // copy member
  cannyEdges&
    cannyEdges::copy(const cannyEdges& other) {
      edgeDetector::copy(other);

    return (*this);
  }

  // alias for copy member
  cannyEdges&
    cannyEdges::operator=(const cannyEdges& other) {
    return (copy(other));
  }


  // clone member
  functor* cannyEdges::clone() const {
    return new cannyEdges(*this);
  }

  // return parameters
  const cannyEdges::parameters&
    cannyEdges::getParameters() const {
    const parameters* par =
      dynamic_cast<const parameters*>(&functor::getParameters());
    if(isNull(par)) {
      throw invalidParametersException(getTypeName());
    }
    return *par;
  }

  // set parameters
  bool cannyEdges::setParameters(const functor::parameters& theParam) {
    if (edgeDetector::setParameters(theParam)) {
      const parameters& par = getParameters();
      nonMaximaSuppression::parameters nmsPar;
      nmsPar.thresholdMin = par.thresholdMin;
      nmsPar.thresholdMax = (par.thresholdMax<=1.0) ? par.thresholdMax :
                                                      par.thresholdMax/255.0f;
      nmsPar.edgeValue    = par.edgeValue;
      nmsPar.background   = par.noEdgeValue;
      nmsPar.checkAngles  = false;
      return suppressor.setParameters(nmsPar);
    }
    return false;
  }

  // -------------------------------------------------------------------
  // The apply-methods!
  // -------------------------------------------------------------------


  // On place apply for type channel8!
  bool cannyEdges::apply(channel8& srcdest) const {
    return edgeDetector::apply(srcdest);
  }

  // On place apply for type channel!
  bool cannyEdges::apply(channel& srcdest) const {
    return edgeDetector::apply(srcdest);
  }

  // On copy apply for type channel8!
  bool cannyEdges::apply(const channel8& src,channel8& dest) const {
    channel orient;
    return apply(src,dest,orient);
  }

  // On copy apply for type channel8!
  bool cannyEdges::apply(const channel8& src,
                               channel8& dest,
                               channel& orientation) const {

    const parameters& param = getParameters();
    if (src.empty()) {
      setStatusString("input image empty");
      return(false);
    }

    dest.resize(src.size(),ubyte(),false,false);

    // temporary images
    channel8 chanSmoothed; // low-pass filtered image (noise reduction)
    channel gradAbs;       // magnitude of gradient

    gradAbs.resize(src.size(),0.0f,false,false);
    orientation.resize(src.size(),0.0f,false,false);

    // apply gauss filter to smooth image
    _lti_debug("Gauss Filter");

    gaussKernel2D<ubyte> kernel(param.kernelSize,param.variance);

    convolution::parameters filterParam;       // parameters
    filterParam.setKernel(kernel);             // use the gauss kernel
    filterParam.boundaryType = lti::Constant;
    convolution filter(filterParam);           // convolution operator

    // filter the channel 
    filter.apply(src,chanSmoothed);

    _lti_debug("Differentiation and result in polar coordinates");

    if (computeGradient(chanSmoothed,gradAbs,orientation)) {

#     ifdef _LTI_DEBUG
      static viewer viewAbs("Abs"),viewArg("Arg");
      viewAbs.show(gradAbs);
      viewArg.show(orientation);
#     endif      

      return suppressor.apply(gradAbs,orientation,dest);
    }
   
    
    return false;
  };

  // On copy apply for type channel8!
  bool cannyEdges::apply(const channel& src,channel8& dest) const {
    channel8 tmp;
    tmp.castFrom(src);
    return apply(tmp,dest);
  }

  // On copy apply for type channel8!
  bool cannyEdges::apply(const channel& src,channel8& dest,channel& o) const {
    channel8 tmp;
    tmp.castFrom(src);
    return apply(tmp,dest,o);
  }


  // On copy apply for type channel!
  bool cannyEdges::apply(const channel& src,channel& dest) const {
    channel8 tmp;
    if (apply(src,tmp)) {
      dest.castFrom(tmp);
      return true;
    }
    return false;
  };

  // On copy apply for type channel!
  bool cannyEdges::apply(const image& src,channel8& dest) const {
    splitImageToRGB splitter;
    channel r,g,b,o;
    splitter.apply(src,r,g,b);
    
    return apply(r,g,b,dest,o);
  };

  // On copy apply for type channel!
  bool cannyEdges::apply(const image& src,channel8& dest,channel& ori) const {
    splitImageToRGB splitter;
    channel r,g,b;
    splitter.apply(src,r,g,b);
    
    return apply(r,g,b,dest,ori);
  };

  // On copy apply for type channel!
  bool cannyEdges::apply(const image& src,image& dest) const {
    return edgeDetector::apply(src,dest);
  };

  // On copy apply for type channel!
  bool cannyEdges::apply(image& srcdest) const {
    return edgeDetector::apply(srcdest);
  };

  bool cannyEdges::apply(const channel& c1,
                         const channel& c2,
                         const channel& c3,
                         channel8& dest,
                         channel& orientation) const {

    const parameters& param = getParameters();

    if ((c1.size() != c2.size()) ||
        (c1.size() != c3.size())) {
      setStatusString("Input channels have different sizes");
      return false;
    }

    if (c1.empty()) {
      setStatusString("input image empty");
      return(false);
    }    

    dest.resize(c1.size(),ubyte(),false,false);

    // temporary images
    channel c1s,c2s,c3s; // low-pass filtered image (noise reduction)
    channel gradAbs;       // magnitude of gradient

    gradAbs.resize(c1.size(),0.0f,false,false);
    orientation.resize(c1.size(),0.0f,false,false);

    // apply gauss filter to smooth image
    _lti_debug("Gauss Filter");

    gaussKernel2D<ubyte> kernel(param.kernelSize,param.variance);

    convolution::parameters filterParam;       // parameters
    filterParam.setKernel(kernel);             // use the gauss kernel
    filterParam.boundaryType = lti::Constant;
    convolution filter(filterParam);           // convolution operator

    // filter the channel 
    filter.apply(c1,c1s);
    filter.apply(c2,c2s);
    filter.apply(c3,c3s);

    _lti_debug("Differentiation and result in polar coordinates");

    float maxVal;
    
    colorContrastGradient colorGrad(colorContrastGradient::parameters::Polar,
                                    param.optimalGradientKernelSize);


    if (colorGrad.apply(c1s,c2s,c3s,gradAbs,orientation,maxVal)) {

#     ifdef _LTI_DEBUG
      static viewer viewAbs("Abs Color"),viewArg("Arg");
      viewAbs.show(gradAbs);
      viewArg.show(orientation);
#     endif      

      return suppressor.apply(gradAbs,orientation,dest,maxVal);
    }   
    
    return false;
  }

  // compute the gradient
  /*
   * compute the gradient of the src channel8 using the adecuate
   * method xyDifferentiateXYZ
   */
  bool cannyEdges::computeGradient(const channel8& src,
                                   channel& gradAbs,
                                   channel& orientation) const {
    
    bool result;
    //XY-Differentiate Image and get gradient in polar coordinates:
    switch(getParameters().optimalGradientKernelSize){
    case(2):
      //inter-pixel gradient
      result=xyDifferentiateImageInterPixel(src,gradAbs,orientation);
      break;
    case(3):
    case(4):
    case(5):
      //the optimal gradient kernels
      result=xyDifferentiateImageOpt(src,gradAbs,orientation);
      break;
    default:
      result=xyDifferentiateImage(src,gradAbs,orientation);
    }
    return result;
  }

  // ----


  bool
  cannyEdges::xyDifferentiateImageCartOpt(const channel& src,
                                          channel& dx,
                                          channel& dy) const {

#   ifdef _LTI_DEBUG
    static viewer viewGX("GX"),viewGY("GY");
#   endif

    const parameters& par = getParameters();

    gradientKernelX<channel::value_type> gx(par.optimalGradientKernelSize);
    gradientKernelY<channel::value_type> gy(par.optimalGradientKernelSize);

    convolution grad;
    convolution::parameters gradpar;

    gradpar.boundaryType = lti::Constant;

    // x direction
    gradpar.setKernel(gx);
    grad.setParameters(gradpar);
    grad.apply(src,dx);

    // y direction
    grad.setKernel(gy);
    grad.apply(src,dy);

#   ifdef _LTI_DEBUG
    viewGX.show(dx);
    viewGY.show(dy);
#   endif      

    return true;
  }

  bool
  cannyEdges::xyDifferentiateImageOpt(const channel8& src,
                                      channel& gradAbs,
                                      channel& orientation) const {

    channel orig,dx,dy;
    orig.castFrom(src);

    xyDifferentiateImageCartOpt(orig,dx,dy);
    // we can assume, all four channels are connected!

    // version with 4 increments and 4 direct accesses
    float *pdx=&dx.at(0,0);
    float *pdy=&dy.at(0,0);
    float *pabs=&gradAbs.at(0,0);
    float *pori=&orientation.at(0,0);
    float *pend;
    
    for (pend=pdx+(dx.rows()*dx.columns());
         pdx<pend;
         ++pdx,++pdy,++pabs,++pori) {
      cartToPolar(iround((*pdx)*255),
                  iround((*pdy)*255),
                  *pabs,*pori);
    }
    return true;
  }

  bool
  cannyEdges::xyDifferentiateImage(const channel8& src, //in:  Source Image
                                   channel& gradAbs,    //out: Magnitude
                                   channel& gradPhi     //out: Orientation
                                   ) const {

    if (src.columns() < 3) {
      setStatusString("width less than 3");
      return false;
    }
    if (src.rows() < 3) {
      setStatusString("height less than 3");
      return false;
    }


    const int width  = src.columns();
    const int height = src.rows();

    float* fpGradAbs = &gradAbs.at(0,0);
    float* fpGradPhi = &gradPhi.at(0,0);

    const ubyte* fpSrc = &src.at(0,0);
    const ubyte* rowy;
    const ubyte* colx;

    float* pidx;
    float* pidy;

    const int w1 = width-1;
    const int w2 = width-2;
    const int last = (height-1)*width; // index of begin of last row
    const int lastRow = -w1;           // offset from actual column pointer to
                                       // last row
    const int nextRow = width+1;       // offset from actual column pointer to
                                       // next row

    // top-left corner
    cartToPolar(fpSrc[1]-fpSrc[0],fpSrc[width]-fpSrc[0],
                fpGradAbs[0],fpGradPhi[0]);

    // top
    pidx = &fpGradAbs[1];
    pidy = &fpGradPhi[1];

    for (colx=&fpSrc[0],rowy=&fpSrc[w1];
         colx<rowy;
         ++colx,++pidx,++pidy) {
      cartToPolar(colx[2] - *colx,colx[nextRow] - colx[1],*pidx,*pidy);
    }

    // top-right corner
    cartToPolar(fpSrc[w1]-fpSrc[w2],fpSrc[width+w1]-fpSrc[w1],
                fpGradAbs[w1],fpGradPhi[w1]);

    // main loop (begin at coordinates (1,0)
    pidx = &fpGradAbs[width];
    pidy = &fpGradPhi[width];

    const ubyte *const rowEnd = &fpSrc[last];

    for (rowy=&fpSrc[width];
         rowy<rowEnd;
         rowy+=width) {

      cartToPolar(rowy[1] - rowy[0],rowy[width] - rowy[-width],*pidx,*pidy);

      ++pidx;
      ++pidy;

      const ubyte *const colEnd = &rowy[w2];
      for (colx=rowy;
           colx<colEnd;
           ++colx,++pidx,++pidy) {
        cartToPolar(colx[2] - *colx,colx[nextRow] - colx[lastRow],*pidx,*pidy);
      }

      cartToPolar(colx[1] - *colx,colx[nextRow] - colx[lastRow],*pidx,*pidy);
      ++pidx;
      ++pidy;
    }

    // bottom-left corner
    cartToPolar(fpSrc[last+1]-fpSrc[last],fpSrc[last]-fpSrc[last-width],
                fpGradAbs[last],fpGradPhi[last]);

    // bottom
    pidx = &fpGradAbs[last+1];
    pidy = &fpGradPhi[last+1];

    const ubyte *const colEnd = &rowEnd[w2];
    for (colx=rowEnd;
         colx<colEnd;
         ++colx,++pidy,++pidx) {
      cartToPolar(colx[2]-*colx,colx[1]-colx[lastRow],*pidx,*pidy);
    }

    // bottom-right corner
    cartToPolar(fpSrc[last+w1]-fpSrc[last+w2],
                fpSrc[last+w1]-fpSrc[last-width+w1],
                fpGradAbs[last+w1],fpGradPhi[last+w1]);

    return true;
  };


  // IntePixel gradient calculation
  bool
  cannyEdges::xyDifferentiateImageInterPixel(const channel8& src,
                                             channel& gradAbs,
                                             channel& orientation) const {

    const int iWidth  = src.columns();
    const int iHeight = src.rows();
    
    int iX,iY;
    float fAbs,fPhi;
    static const float correctAngle = static_cast<float>(0.25*Pi);

    // main block
    for (iY=0; iY< src.lastRow(); iY++) {
      for (iX=0; iX< src.lastColumn(); iX++) {
        cartToPolar(src.at(iY+1,iX+1) - src.at(iY,iX),
                    -src.at(iY,iX+1)  + src.at(iY+1,iX),fAbs,fPhi);
        fPhi+=correctAngle;
        if (fPhi>constants<float>::TwicePi) fPhi-=constants<float>::TwicePi;
        gradAbs.at(iY,iX)=fAbs;
        orientation.at(iY,iX)=fPhi;
      }
    }
    
    // last column
    for (iY=0; iY< iHeight-1; iY++) {
      iX=iWidth-1;
      cartToPolar(src.at(iY+1,iX)-src.at(iY,iX),
                  -src.at(iY,iX)+src.at(iY+1,iX),fAbs,fPhi);
      fPhi+=correctAngle;
      if (fPhi>constants<float>::TwicePi) fPhi+=constants<float>::TwicePi;
      gradAbs.at(iY,iX)=fAbs;
      orientation.at(iY,iX)=fPhi;
    }
    
    // last row
    for (iX=0; iX< iWidth-1; iX++){
      iY=iHeight-1;
      cartToPolar(src.at(iY,iX+1)-src.at(iY,iX), 
                  -src.at(iY,iX+1)+src.at(iY,iX),fAbs,fPhi);
      fPhi+=correctAngle;
      if (fPhi>constants<float>::TwicePi) fPhi+=constants<float>::TwicePi;
      gradAbs.at(iY,iX)=fAbs;
      orientation.at(iY,iX)=fPhi;
    }
    gradAbs.at(iHeight-1,iWidth-1)=0.0;
    orientation.at(iHeight-1,iWidth-1)=0.0;

    return true;
  }
}
