/*
 * Copyright (C) 1998, 1999, 2000, 2001
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 *
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.
 */


/* -------------------------------------------------------------------
 * project ....: LTI-Lib: Image Processing and Computer Vision Library
 * file .......: ltiColorEdgesGS.cpp
 * authors ....: Pablo Alvarado
 * organization: LTI, RWTH Aachen
 * creation ...: 17.10.2003
 * revisions ..: $Id: ltiColorEdgesGS.cpp,v 1.4 2003/11/05 10:38:25 alvarado Exp $
 */

#include "ltiColorEdgesGS.h"
#include "ltiArctanLUT.h"
#include "ltiGradientKernels.h"
#include "ltiOgdKernels.h"
#include "ltiConvolution.h"

#undef _LTI_DEBUG
// #define _LTI_DEBUG 2
#include "ltiDebug.h"

#ifdef _LTI_DEBUG
#include "ltiViewer.h"
#endif


namespace lti {
  // --------------------------------------------------
  // colorEdgesGS::parameters
  // --------------------------------------------------

  // default constructor
  colorEdgesGS::parameters::parameters()
    : edgeDetector::parameters() {

    shadowEdgeValue = ubyte(128);
    highlightEdgeValue = ubyte(192);
    noiseStdDeviation = trgbPixel<float>(1.5f,1.5f,1.5f);
    gradient.format = colorContrastGradient::parameters::Cartesic;
    uncertaintyFactor = 3.0f;
  }

  // copy constructor
  colorEdgesGS::parameters::parameters(const parameters& other)
    : edgeDetector::parameters() {
    copy(other);
  }

  // destructor
  colorEdgesGS::parameters::~parameters() {
  }

  // get type name
  const char* colorEdgesGS::parameters::getTypeName() const {
    return "colorEdgesGS::parameters";
  }

  // copy member

  colorEdgesGS::parameters&
    colorEdgesGS::parameters::copy(const parameters& other) {
# ifndef _LTI_MSC_VER
    // MS Visual C++ 6 is not able to compile this...
    edgeDetector::parameters::copy(other);
# else
    // ...so we have to use this workaround.
    // Conditional on that, copy may not be virtual.
    edgeDetector::parameters& (edgeDetector::parameters::* p_copy)
      (const edgeDetector::parameters&) =
      edgeDetector::parameters::copy;
    (this->*p_copy)(other);
# endif

    
    shadowEdgeValue = other.shadowEdgeValue;
    highlightEdgeValue = other.highlightEdgeValue;
    noiseStdDeviation = other.noiseStdDeviation;
    gradient.copy(other.gradient);
    uncertaintyFactor = other.uncertaintyFactor;

    return *this;
  }

  // alias for copy member
  colorEdgesGS::parameters&
    colorEdgesGS::parameters::operator=(const parameters& other) {
    return copy(other);
  }

  // clone member
  functor::parameters* colorEdgesGS::parameters::clone() const {
    return new parameters(*this);
  }

  /*
   * write the parameters in the given ioHandler
   * @param handler the ioHandler to be used
   * @param complete if true (the default) the enclosing begin/end will
   *        be also written, otherwise only the data block will be written.
   * @return true if write was succeful
   */
# ifndef _LTI_MSC_VER
  bool colorEdgesGS::parameters::write(ioHandler& handler,
                                         const bool& complete) const
# else
  bool colorEdgesGS::parameters::writeMS(ioHandler& handler,
                                           const bool& complete) const
# endif
  {
    bool b = true;
    if (complete) {
      b = handler.writeBegin();
    }

    if (b) {
      lti::write(handler,"shadowEdgeValue",shadowEdgeValue);
      lti::write(handler,"highlightEdgeValue",highlightEdgeValue);
      lti::write(handler,"noiseStdDeviation",noiseStdDeviation);
      lti::write(handler,"gradient",gradient);
      lti::write(handler,"uncertaintyFactor",uncertaintyFactor);
    }

# ifndef _LTI_MSC_VER
    // This is the standard C++ code, which MS Visual C++ 6 is not able to
    // compile...
    b = b && edgeDetector::parameters::write(handler,false);
# else
    bool (edgeDetector::parameters::* p_writeMS)(ioHandler&,const bool&) const =
      edgeDetector::parameters::writeMS;
    b = b && (this->*p_writeMS)(handler,false);
# endif

    if (complete) {
      b = b && handler.writeEnd();
    }

    return b;
  }

# ifdef _LTI_MSC_VER
  bool colorEdgesGS::parameters::write(ioHandler& handler,
                                         const bool& complete) const {
    // ...we need this workaround to cope with another really awful MSVC bug.
    return writeMS(handler,complete);
  }
# endif

  /*
   * read the parameters from the given ioHandler
   * @param handler the ioHandler to be used
   * @param complete if true (the default) the enclosing begin/end will
   *        be also read, otherwise only the data block will be read.
   * @return true if write was succeful
   */
# ifndef _LTI_MSC_VER
  bool colorEdgesGS::parameters::read(ioHandler& handler,
                                        const bool& complete)
# else
  bool colorEdgesGS::parameters::readMS(ioHandler& handler,
                                          const bool& complete)
# endif
  {
    bool b = true;
    if (complete) {
      b = handler.readBegin();
    }

    if (b) {
      lti::read(handler,"shadowEdgeValue",shadowEdgeValue);
      lti::read(handler,"highlightEdgeValue",highlightEdgeValue);
      lti::read(handler,"noiseStdDeviation",noiseStdDeviation);
      lti::read(handler,"gradient",gradient);
      lti::read(handler,"uncertaintyFactor",uncertaintyFactor);
    }

# ifndef _LTI_MSC_VER
    // This is the standard C++ code, which MS Visual C++ 6 is not able to
    // compile...
    b = b && edgeDetector::parameters::read(handler,false);
# else
    bool (edgeDetector::parameters::* p_readMS)(ioHandler&,const bool&) =
      edgeDetector::parameters::readMS;
    b = b && (this->*p_readMS)(handler,false);
# endif

    if (complete) {
      b = b && handler.readEnd();
    }

    return b;
  }

# ifdef _LTI_MSC_VER
  bool colorEdgesGS::parameters::read(ioHandler& handler,
                                        const bool& complete) {
    // ...we need this workaround to cope with another really awful MSVC
    // bug.
    return readMS(handler,complete);
  }
# endif

  bool colorEdgesGS::parameters::measureNoise(const image& img,
                                              const rectangle& window) {
    int minX,maxX,minY,maxY;
    minX = max(0,min(window.ul.x,window.br.x));
    minY = max(0,min(window.ul.y,window.br.y));

    maxX = min(img.lastColumn(),max(window.ul.x,window.br.x));
    maxY = min(img.lastRow(),max(window.ul.y,window.br.y));

    int y,x;
    trgbPixel<float> sum(0,0,0),sum2(0,0,0);
    trgbPixel<float> c;
    int n = 0;

    for (y=minY;y<=maxY;++y) {
      for (x=minX;x<=maxX;++x) {
        c.copy(img.at(y,x));
        sum.add(c);
        sum2.add(trgbPixel<float>(c.red*c.red,c.green*c.green,c.blue*c.blue));
        n++;
      }
    }
    sum.divide(static_cast<float>(n));
    sum2.divide(static_cast<float>(n));
    noiseStdDeviation.set(sqrtrect(sum2.red   - sum.red*sum.red),
                          sqrtrect(sum2.green - sum.green*sum.green),
                          sqrtrect(sum2.blue  - sum.blue*sum.blue));
    
    return true;
  }

  // --------------------------------------------------
  // colorEdgesGS
  // --------------------------------------------------

  // default constructor
  colorEdgesGS::colorEdgesGS()
    : edgeDetector(){

    // create an instance of the parameters with the default values
    parameters defaultParameters;
    // set the default parameters
    setParameters(defaultParameters);
  }

  // default constructor
  colorEdgesGS::colorEdgesGS(const parameters& par)
    : edgeDetector() {

    // set the given parameters
    setParameters(par);
  }


  // copy constructor
  colorEdgesGS::colorEdgesGS(const colorEdgesGS& other)
    : edgeDetector() {
    copy(other);
  }

  // destructor
  colorEdgesGS::~colorEdgesGS() {
  }

  // returns the name of this type
  const char* colorEdgesGS::getTypeName() const {
    return "colorEdgesGS";
  }

  // which apply is faster
  functor::eFasterApply colorEdgesGS::theFaster() const {
    return OnCopy;
  }

  // copy member
  colorEdgesGS&
  colorEdgesGS::copy(const colorEdgesGS& other) {
    edgeDetector::copy(other);
      
    return (*this);
  }

  // alias for copy member
  colorEdgesGS&
    colorEdgesGS::operator=(const colorEdgesGS& other) {
    return (copy(other));
  }


  // clone member
  functor* colorEdgesGS::clone() const {
    return new colorEdgesGS(*this);
  }

  // return parameters
  const colorEdgesGS::parameters& colorEdgesGS::getParameters() const {
    const parameters* par =
      dynamic_cast<const parameters*>(&functor::getParameters());
    if(isNull(par)) {
      throw invalidParametersException(getTypeName());
    }
    return *par;
  }

  // -------------------------------------------------------------------
  // The apply-methods!
  // -------------------------------------------------------------------

  bool colorEdgesGS::apply(channel8 &srcdest) const {
    image tmp;
    tmp.castFrom(srcdest);
    return apply(tmp,srcdest);
  }

  bool colorEdgesGS::apply(channel &srcdest) const {
    image tmp;
    channel8 chnl8;

    tmp.castFrom(srcdest);
    if (apply(tmp,chnl8)) {
      srcdest.castFrom(chnl8);
      return true;
    }
    
    return false;
  }

  bool colorEdgesGS::apply(const channel8 &src, channel8 &dest) const {
    image tmp;
    tmp.castFrom(src);
    return apply(tmp,dest);    
  }

  bool colorEdgesGS::apply(const channel &src, channel &dest) const {
    image tmp;
    channel8 chnl8;

    tmp.castFrom(src);
    if (apply(tmp,chnl8)) {
      dest.castFrom(chnl8);
      return true;
    }
    
    return false;
  } 

  bool colorEdgesGS::apply(const channel &src, channel8 &dest) const {
    image tmp;
    tmp.castFrom(src);
    return apply(tmp,dest);
  } 

  bool colorEdgesGS::apply (image &srcdest) const {
    image tmp;
    if (apply(srcdest,tmp)) {
      tmp.detach(srcdest);
      return true;
    }
    return false;
  }

  bool colorEdgesGS::apply (const image &src, image &dest) const {
    const parameters& par = getParameters();
    const rgbPixel back = rgbPixel(par.noEdgeValue,
                                   par.noEdgeValue,
                                   par.noEdgeValue);
    const rgbPixel mate = rgbPixel(par.edgeValue,
                                   par.edgeValue,
                                   par.edgeValue);

    const rgbPixel shad = rgbPixel(0,par.edgeValue,0);
    const rgbPixel high = rgbPixel(0,par.edgeValue,0);
    
    channel8 mask;
    if (apply(src,mask)) {
      dest.resize(src.size(),rgbPixel(),false,false);
      channel8::const_iterator mit;
      image::iterator it,eit;
      eit=dest.end();

      for (it=dest.begin(),mit=mask.begin();it!=eit;++it,++mit) {
        if ((*mit) == par.noEdgeValue) {
          (*it)=back;
        } else if ((*mit) == par.edgeValue) {
          (*it)=mate;
        } else if ((*mit) == par.shadowEdgeValue) {
          (*it)=shad;
        } else if ((*mit) == par.highlightEdgeValue) {
          (*it)=high;
        } else {
          // this shouldn't happend, but to avoid uninitialized memory:
          (*it) = lti::Black;
        }
      }
      return true;
    }

    return false;
  }

  bool colorEdgesGS::apply (const image &src, channel8 &dest) const {
    if (src.empty()) {
      dest.clear();
      return true;
    }

    const parameters& par = getParameters();

    channel red,green,blue;
    channel o1,o2,c1,c2;
    trgbPixel<float> stdDevRGB;
    channel stdDevC1,stdDevC2;
    channel rgbGrad,rgbUnc,ocGrad,ocUnc,cGrad,cUnc;
    float stdDevO1;
    float stdDevO2;


    getChannels(src,red,green,blue,o1,o2,c1,c2,stdDevRGB,stdDevO1,stdDevO2,
                stdDevC1,stdDevC2);

    getGradients(red,green,blue,o1,o2,c1,c2,stdDevRGB,stdDevO1,stdDevO2,
                 stdDevC1,stdDevC2,
                 rgbGrad,rgbUnc,ocGrad,ocUnc,cGrad,cUnc);

    dest.resize(src.size(),ubyte(),false,false);

    int y,x;
    bool drgb,dcs,dos;
    for (y=0;y<rgbGrad.rows();++y) {
      for (x=0;x<rgbGrad.columns();++x) {
        if (rgbGrad.at(y,x) < rgbUnc.at(y,x)) {
          rgbGrad.at(y,x)=0.0f;
          drgb = false;
        } else {
          drgb = true;
        }

        if (ocGrad.at(y,x) < ocUnc.at(y,x)) {
          ocGrad.at(y,x)=0.0f;
          dos = false;
        } else {
          dos = true;
        }

        if (cGrad.at(y,x) < cUnc.at(y,x)) {
          cGrad.at(y,x)=0.0f;
          dcs = false;
        } else {
          dcs = true;
        }

        if (drgb) {
          if (!dcs) {
            dest.at(y,x) = par.shadowEdgeValue;
          } else if (dcs && !dos) {
            dest.at(y,x) = par.highlightEdgeValue;
          } else {
            dest.at(y,x) = par.edgeValue;
          }
        } else {
          dest.at(y,x) = par.noEdgeValue;
        }
      }
    }
    
#ifdef _LTI_DEBUG 
    static viewer v1("rgb");
    static viewer v2("rgb u");
    static viewer v3("oc");
    static viewer v4("oc u");
    static viewer v5("c");
    static viewer v6("c u");
    static viewer ve("edges");

    ve.show(dest);

    v1.show(rgbGrad);
    v2.show(rgbUnc);

    v3.show(ocGrad);
    v4.show(ocUnc);

    v5.show(cGrad);
    v6.show(cUnc);
    
#endif

    return true;
  }

  bool colorEdgesGS::getChannels(const image& img,
                                 channel& red,
                                 channel& green,
                                 channel& blue,
                                 channel& o1,
                                 channel& o2,
                                 channel& c1,
                                 channel& c2,
                                 trgbPixel<float>& stdDevRGB,
                                 float& stdDevO1,
                                 float& stdDevO2,
                                 channel& stdDevC1,
                                 channel& stdDevC2) const {
    
    // compute the constant standard deviations
    const parameters& par = getParameters();

    stdDevRGB.divide(par.noiseStdDeviation,255.0f);

    trgbPixel<float> varRGB(stdDevRGB.red*stdDevRGB.red,
                            stdDevRGB.green*stdDevRGB.green,
                            stdDevRGB.blue*stdDevRGB.blue);

    stdDevO1 = 0.5f*sqrt(varRGB.green + varRGB.red);
    stdDevO2 = 0.5f*sqrt(4.0f*varRGB.blue + varRGB.green + varRGB.red);

    /*
     * Compute color channels
     */

    red.resize(img.size(),float(),false,false);
    green.resize(img.size(),float(),false,false);
    blue.resize(img.size(),float(),false,false);
    o1.resize(img.size(),float(),false,false);
    o2.resize(img.size(),float(),false,false);
    c1.resize(img.size(),float(),false,false);
    c2.resize(img.size(),float(),false,false);

    stdDevC1.resize(img.size(),float(),false,false);
    stdDevC2.resize(img.size(),float(),false,false);

    arctanLUT atan2;

    int y,x;
    ubyte r,g,b;
    float fr,fg,fb;

    for (y=0;y<img.rows();++y) {
      for (x=0;x<img.columns();++x) {
        const rgbPixel& px = img.at(y,x);
        r=px.getRed();
        g=px.getGreen();
        b=px.getBlue();

        fr = red.at(y,x)   = float(r)/255.0f;
        fg = green.at(y,x) = float(g)/255.0f;
        fb = blue.at(y,x)  = float(b)/255.0f;

        c1.at(y,x) = atan2(r,b);
        c2.at(y,x) = atan2(g,b);

        o1.at(y,x) = 0.5f*(fr-fg);
        o2.at(y,x) = 0.5f*fb - 0.25f*(fr+fg);

        if (b==0) {
          if (r == 0) {
            stdDevC1.at(y,x) = 0.0f;
          } else {
            stdDevC1.at(y,x) = sqrt(fr*fr*varRGB.blue)/(fr*fr);
          }

          if (g == 0) {
            stdDevC2.at(y,x) = 0.0f;
          } else {
            stdDevC2.at(y,x) = sqrt(fg*fg*varRGB.blue)/(fg*fg);
          }
        } else {
          stdDevC1.at(y,x) = sqrt(fr*fr*varRGB.blue + fb*varRGB.red)/
                             (fr*fr + fb*fb);
          stdDevC2.at(y,x) = sqrt(fg*fg*varRGB.blue + fb*varRGB.green)/
                             (fg*fg + fb*fb);
        }
      }
    }

    return true;
  }
  

  bool colorEdgesGS::getGradients(const channel& red,
                                  const channel& green,
                                  const channel& blue,
                                  const channel& o1,
                                  const channel& o2,
                                  const channel& c1,
                                  const channel& c2,
                                  const trgbPixel<float> &stdDevRGB,
                                  const float& stdDevO1,
                                  const float& stdDevO2,
                                  const channel& stdDevC1,
                                  const channel& stdDevC2,
                                  channel& rgbGrad,
                                  channel& rgbUnc,
                                  channel& ocGrad,
                                  channel& ocUnc,
                                  channel& cGrad,
                                  channel& cUnc) const {
    

    const parameters& par = getParameters();
    int y,x;

    channel gx1,gy1,gx2,gy2,gx3,gy3;

    colorContrastGradient::parameters ccgPar(par.gradient);
    ccgPar.format = colorContrastGradient::parameters::Cartesic;
    colorContrastGradient ccg(ccgPar);

    sepKernel<float> ux,uy;
    constructUncertaintyKernels(ux,uy);

    kernel2D<float> ux2,uy2;
    ux2.castFrom(ux);
    uy2.castFrom(uy);

    const float uvx = ux2.sumOfElements();
    const float uvy = uy2.sumOfElements();

    // ------------------------
    // the RGB gradient modulus
    // ------------------------

    ccg.apply(red,gx1,gy1);
    ccg.apply(green,gx2,gy2);
    ccg.apply(blue,gx3,gy3);

    rgbGrad.resize(red.size(),float(),false,false);
    rgbUnc.resize(red.size(),float(),false,false);

    // these are the sigma_(dci/dx) and sigma_(dci/dy) for i = r,g and b
    const float sdrx = stdDevRGB.red*uvx;
    const float sdry = stdDevRGB.red*uvy;

    const float sdgx = stdDevRGB.green*uvx;
    const float sdgy = stdDevRGB.green*uvy;

    const float sdbx = stdDevRGB.blue*uvx;
    const float sdby = stdDevRGB.blue*uvy;

    float mod,det;
    float fgx1,fgx2,fgx3,fgy1,fgy2,fgy3;

    for (y=0;y<rgbGrad.rows();++y) {
      for (x=0;x<rgbGrad.columns();++x) {
        fgx1 = gx1.at(y,x);
        fgy1 = gy1.at(y,x);
        fgx2 = gx2.at(y,x);
        fgy2 = gy2.at(y,x);
        fgx3 = gx3.at(y,x);
        fgy3 = gy3.at(y,x);

        mod = sqrt(fgx1*fgx1 + fgx2*fgx2 + fgx3*fgx3 +
                   fgy1*fgy1 + fgy2*fgy2 + fgy3*fgy3);

        det = abs(fgx1)*sdrx + abs(fgx2)*sdgx + abs(fgx3)*sdbx +
              abs(fgy1)*sdry + abs(fgy2)*sdgy + abs(fgy3)*sdby;

        rgbGrad.at(y,x) = mod;
        rgbUnc.at(y,x) = (mod != 0.0) ? par.uncertaintyFactor*det/mod : mod;
      }
    }

    // -------------------------
    // the o1o2 gradient modulus
    // -------------------------

    // these are the sigma_(dci/dx) and sigma_(dci/dy) for i = r,g and b
    const float sdo1x = stdDevO1*uvx;
    const float sdo1y = stdDevO1*uvy;

    const float sdo2x = stdDevO2*uvx;
    const float sdo2y = stdDevO2*uvy;

    ocGrad.resize(red.size(),float(),false,false);
    ocUnc.resize(red.size(),float(),false,false);

    ccg.apply(o1,gx1,gy1);
    ccg.apply(o2,gx2,gy2);

    for (y=0;y<rgbGrad.rows();++y) {
      for (x=0;x<rgbGrad.columns();++x) {

        fgx1 = gx1.at(y,x);
        fgy1 = gy1.at(y,x);
        fgx2 = gx2.at(y,x);
        fgy2 = gy2.at(y,x);

        mod = sqrt(fgx1*fgx1 + fgx2*fgx2 +
                   fgy1*fgy1 + fgy2*fgy2);

        det = abs(fgx1)*sdo1x + abs(fgx2)*sdo2x +
              abs(fgy1)*sdo1y + abs(fgy2)*sdo2y;

        ocGrad.at(y,x) = mod;
        ocUnc.at(y,x) = (mod != 0.0) ? par.uncertaintyFactor*det/mod : mod;
      }
    }

    // -------------------------
    // the c1c2 gradient modulus
    // -------------------------

    cGrad.resize(red.size(),float(),false,false);
    cUnc.resize(red.size(),float(),false,false);

    // these are the sigma_(dci/dx) and sigma_(dci/dy) for i = r,g and b
    ccg.apply(c1,gx1,gy1);
    ccg.apply(c2,gx2,gy2);

    convolution convPar(ux,convolution::parameters::Constant);
    channel uc1x,uc1y,uc2x,uc2y;

    convPar.apply(stdDevC1,uc1x);
    convPar.apply(stdDevC2,uc2x);

    convPar.setKernel(uy);
    convPar.apply(stdDevC1,uc1y);
    convPar.apply(stdDevC2,uc2y);

    for (y=0;y<rgbGrad.rows();++y) {
      for (x=0;x<rgbGrad.columns();++x) {

        fgx1 = gx1.at(y,x);
        fgy1 = gy1.at(y,x);
        fgx2 = gx2.at(y,x);
        fgy2 = gy2.at(y,x);

        mod = sqrt(fgx1*fgx1 + fgx2*fgx2 +
                   fgy1*fgy1 + fgy2*fgy2);

        det = abs(fgx1)*uc1x.at(y,x) + abs(fgx2)*uc2x.at(y,x) +
              abs(fgy1)*uc1y.at(y,x) + abs(fgy2)*uc2y.at(y,x);

        cGrad.at(y,x) = mod;
        cUnc.at(y,x) = (mod != 0.0) ? par.uncertaintyFactor*det/mod : mod;
      }
    }

    return true;
  }

  bool 
  colorEdgesGS::constructUncertaintyKernels(sepKernel<float>& filterX,
                                            sepKernel<float>& filterY) const {
    const parameters& par = getParameters();
    const colorContrastGradient::parameters& ccgPar = par.gradient;
    
    if (ccgPar.useOGDKernel) {
      ogd1Kernel<float> gx(ccgPar.optimalGradientKernelSize,
                           ccgPar.ogdVariance,
                           0.0);
      ogd1Kernel<float> gy(ccgPar.optimalGradientKernelSize,
                           ccgPar.ogdVariance,
                           Pi/2.0);

      filterX.apply(gx,abs);
      filterY.apply(gy,abs);
    } else {
      switch (ccgPar.optimalGradientKernelSize) {
        case 0:
        case 1:
        case 2:
          filterX.setNumberOfPairs(1);
          filterX.getRowFilter(0).resize(-1,1,1,false,true);
          filterX.getRowFilter(0).at(0)=0.0f;
          filterX.getColFilter(0).resize(0,0,1,false,true);

          filterY.setNumberOfPairs(1);
          filterY.getColFilter(0).resize(-1,1,1,false,true);
          filterY.getColFilter(0).at(0)=0.0f;
          filterY.getRowFilter(0).resize(0,0,1,false,true);
          
          break;
        case 3:
        case 4:
        case 5:
          gradientKernelX<float> gx(ccgPar.optimalGradientKernelSize);
          gradientKernelY<float> gy(ccgPar.optimalGradientKernelSize);

          filterX.apply(gx,abs);
          filterY.apply(gy,abs);

          break;
      }
    }
    
    return true;
  }

}

