/*
 * Copyright (C) 1998, 1999, 2000, 2001
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 *
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.
 */


/* -----------------------------------------------------------------------
 * project ....: LTI Digitale Bild/Signal Verarbeitungsbibliothek
 * file .......: ltiConvolutionHelper_template.h
 * authors ....: Pablo Alvarado
 * organization: LTI, RWTH Aachen
 * creation ...: 17.11.99
 * revisions ..: $Id: ltiConvolutionHelper_template.h,v 1.19 2003/05/05 06:15:20 alvarado Exp $
 */

namespace lti {

  // ---------------------------------------------
  //   CONVOLUTION HELPER CLASSES
  // ---------------------------------------------


  // ---------------------------------------------
  // 1D DILATION
  // ---------------------------------------------

  template<class T,class A>
    lti::convHelper1D<T,A>::convHelper1D()
    : kernel(0),tmpKernel(0) {
  }

  template<class T,class A>
    lti::convHelper1D<T,A>::~convHelper1D() {
    delete tmpKernel;
  }

  template<class T,class A>
    bool lti::convHelper1D<T,A>::setKernel(const mathObject* kern) {

    kernel = dynamic_cast<const kernel1D<T>*>(kern);

    // it is not the expected type, try to cast from know possibilities
    if (isNull(kernel)) {

      delete tmpKernel;
      tmpKernel = 0;

      // may be a kernel of integers?
      if (notNull(dynamic_cast<const ikernel1D* >(kern))) {
        tmpKernel = new kernel1D<T>;
        tmpKernel->castFrom(*dynamic_cast<const ikernel1D*>(kern));
        if (typeInfo<T>::isFloatingPointType()) {
          tmpKernel->denormalize();
        }
        kernel = tmpKernel;
      } else if (notNull(dynamic_cast<const bkernel1D*>(kern))) {
        tmpKernel = new kernel1D<T>;
        tmpKernel->castFrom(*dynamic_cast<const bkernel1D*>(kern));
        if (typeInfo<T>::isFloatingPointType()) {
          tmpKernel->denormalize();
        }
        kernel = tmpKernel;
      } else if (notNull(dynamic_cast<const fkernel1D*>(kern))) {
        tmpKernel = new kernel1D<T>;

        if (typeInfo<T>::isFloatingPointType()) {
          tmpKernel->castFrom(*dynamic_cast<const fkernel1D*>(kern));
        } else {
          fkernel1D* tmpFloat =
            new fkernel1D(*dynamic_cast<const fkernel1D*>(kern));
          tmpFloat->multiply(static_cast<float>(typeInfo<T>::suggestedNorm()));
          tmpKernel->castFrom(*tmpFloat);
          delete tmpFloat;
        }

        kernel = tmpKernel;
      } else if (notNull(dynamic_cast<const dkernel1D*>(kern))) {
        tmpKernel = new kernel1D<T>;

        if (typeInfo<T>::isFloatingPointType()) {
          tmpKernel->castFrom(*dynamic_cast<const dkernel1D*>(kern));
        } else {
          dkernel1D* tmpFloat =
            new dkernel1D(*dynamic_cast<const dkernel1D*>(kern));
          tmpFloat->multiply(static_cast<double>(typeInfo<T>::suggestedNorm()));
          tmpKernel->castFrom(*tmpFloat);
          delete tmpFloat;
        }

        kernel = tmpKernel;
      }

    }

    return (notNull(kernel));
  }

  template<class T,class A>
  void lti::convHelper1D<T,A>::apply(const vector<T>& src,
                                           vector<T>& dest,
                                     const modifier::parameters& param) {

    int x,i,f,xx,srcSize2;
    int begin,end;
    lresult.reset();

    const T *xp,*fp,*xxp,*filterLastp,*filterFirstp,*endp;
    T *ip;

    // alias for kernel term
    const kernel1D<T>& kern = *kernel;

    // initialize member "norm"
    lresult.setNorm( kern.getNorm() );

    // need kernel as vector
    const vector<T>& filter = kern;

    // last valid index of the filter vector
    const int& filterLast = filter.lastIdx();

    // initialize size of dest (not need to initialization!)
    if (dest.size() != src.size()) {
      dest.resize(src.size(),T(0),false,false);
    }

    // middle block filter (the whole kernel can be applied here!)
    begin = (kern.lastIdx()<0) ? -kern.lastIdx() : 0;
    end   = (kern.lastIdx()>=filter.size()) ?
      src.size()-kern.lastIdx() :
      src.size()-filter.size()+1;

    // This is the standard way to implement the dilation:
    //   for (x=begin,i=kern.lastIdx()+x;x<end;++x,++i) {
    //     // x: begin in src vector
    //     // i: place to write local dilation
    //     // lresult : local result
    //     // xx: sample on image to be checked
    //     for (xx=x,lresult.reset(),f=filterLast;f>=0;--f,++xx ) {
    //       lresult.accumulate(filter.at(f),src.at(xx));
    //     }
    //     dest.at(i)=lresult.getResult();
    //   }

    endp = &src.at(0);
    endp+=end;

    filterFirstp= &filter.at(0);
    --filterFirstp;
    filterLastp = &filter.at(filterLast);

    if (src.size() >= filter.size()) {
      // ----------------------------------------
      // if the kernel is smaller than the vector

      // The dilation with pointer arithmetic (faster this way!)
      for (xp=&src.at(begin),ip=&dest.at(kern.lastIdx()+begin);
           xp!=endp;++xp,++ip) {
        // x: begin in src vector
        // i: place to write local dilation
        // lresult : local result
        // xx: sample on image to be checked
        for (xxp=xp,lresult.reset(),fp=filterLastp;
             fp!=filterFirstp;--fp,++xxp ) {
          lresult.accumulate(*fp,*xxp);
        }
        (*ip)=lresult.getResult();
      }

      // -----------------------
      //    border conditions
      // -----------------------

      switch (param.boundaryType) {
        case lti::NoBoundary:
          break;
        case lti::Zero:
          // left border!
          for (x=begin-1,i=kern.lastIdx()+x;
               x>=-kern.lastIdx();
               --x,--i) {
            for (xx=x+filterLast,lresult.reset(),f=0;
                 ((xx>=0) && (f<filter.size()));
                 ++f,--xx) {
              lresult.accumulate(filter.at(f),src.at(xx));
            }
            dest.at(i)=lresult.getResult();
          }

          // right border!
          for (x=end,i=kern.lastIdx()+x;
               x<src.size()-kern.lastIdx();
               ++x,++i) {
            for (xx=x,lresult.reset(),f=filterLast;
                 ((xx<src.size()) && (f>=0));
                 --f,++xx) {
              lresult.accumulate(filter.at(f),src.at(xx));
            }
            dest.at(i)=lresult.getResult();
          }

          break;

        case lti::Mirror:
          // left border!
          for (x=begin-1,i=kern.lastIdx()+x;
               x>=-kern.lastIdx();
               --x,--i) {
            for (xx=x+filterLast,lresult.reset(),f=0;
                 (f<filter.size());
                 ++f,--xx) {
              if (xx>=0)
                lresult.accumulate(filter.at(f),src.at(xx));
              else
                lresult.accumulate(filter.at(f),src.at(-xx));
            }
            dest.at(i)=lresult.getResult();
          }

          // right border!
          srcSize2=2*src.size()-1;
          for (x=end,i=kern.lastIdx()+x;
               x<src.size()-kern.lastIdx();
               ++x,++i) {
            for (xx=x,lresult.reset(),f=filterLast;
                 (f>=0);
                 --f,++xx) {
              if (xx<src.size())
                lresult.accumulate(filter.at(f),src.at(xx));
              else
                lresult.accumulate(filter.at(f),src.at(srcSize2-xx));
            }
            dest.at(i)=lresult.getResult();
          }
          break;

        case lti::Periodic:
          // left border!
          for (x=begin-1,i=kern.lastIdx()+x;
               x>=-kern.lastIdx();
               --x,--i) {
            for (xx=x+filterLast,lresult.reset(),f=0;
                 (f<filter.size());
                 ++f,--xx) {
              if (xx>=0)
                lresult.accumulate(filter.at(f),src.at(xx));
              else
                lresult.accumulate(filter.at(f),src.at(src.size()+xx));
            }
            dest.at(i)=lresult.getResult();
          }

          // right border!
          for (x=end,i=kern.lastIdx()+x;
               x<src.size()-kern.lastIdx();
               ++x,++i) {
            for (xx=x,lresult.reset(),f=filterLast;
                 (f>=0);
                 --f,++xx) {
              if (xx<src.size())
                lresult.accumulate(filter.at(f),src.at(xx));
              else
                lresult.accumulate(filter.at(f),src.at(xx-src.size()));
            }
            dest.at(i)=lresult.getResult();
          }

          break;
        case lti::Constant:
          T y;

          y = src.at(0);
          // left border!
          for (x=begin-1,i=kern.lastIdx()+x;
               x>=-kern.lastIdx();
               --x,--i) {
            for (xx=x+filterLast,lresult.reset(),f=0;
                 (f<filter.size());
                 ++f,--xx) {
              if (xx>=0)
                lresult.accumulate(filter.at(f),src.at(xx));
              else
                lresult.accumulate(filter.at(f),y);
            }
            dest.at(i)=lresult.getResult();
          }

          y = src.at(src.lastIdx());
          // right border!
          for (x=end,i=kern.lastIdx()+x;
               x<src.size()-kern.lastIdx();
               ++x,++i) {
            for (xx=x,lresult.reset(),f=filterLast;
                 (f>=0);
                 --f,++xx) {
              if (xx<src.size())
                lresult.accumulate(filter.at(f),src.at(xx));
              else
                lresult.accumulate(filter.at(f),y);
            }
            dest.at(i)=lresult.getResult();
          }

          break;
        default:
          throw exception("unknown border type!");
      }
    } else {
      // ---------------------------------------
      // if the kernel is bigger than the vector

      switch (param.boundaryType) {
        case lti::NoBoundary:
          break;
        case lti::Zero:

          for (x=0;x<src.size();++x) {
            for (f = kern.lastIdx(), xx = x+f,lresult.reset();
                 f >= kern.firstIdx(); --f,--xx) {
              if ((xx>=0) && (xx<src.size())) {
                lresult.accumulate(kern.at(f),src.at(xx));
              }
            }
            dest.at(x) = lresult.getResult();
          }

          break;
        case lti::Mirror:

          for (x=0;x<src.size();++x) {
            for (f = kern.lastIdx(), xx = x+f,lresult.reset();
                 f >= kern.firstIdx(); --f,--xx) {
              if ((xx>=0) && (xx<src.size())) {
                lresult.accumulate(kern.at(f),src.at(xx));
              } else {
                int p,xp;
                p = xx/src.size();
                xp = xx%src.size();

                if (xx < 0) {
                  p = -p + 1;
                  if (xp < 0) {
                      xp += src.size();
                  }
                }

                if ((p & 01) != 0) {
                  xp = src.size()-1-xp;
                }

                lresult.accumulate(kern.at(f),src.at(xp));

              }
            }
            dest.at(x) = lresult.getResult();
          }

          break;
        case lti::Periodic:
          for (x=0;x<src.size();++x) {
            for (f = kern.lastIdx(), xx = x+f ,lresult.reset();
                 f >= kern.firstIdx(); --f,--xx) {
              int xp;

              xp = xx%src.size();

              if (xp < 0) {
                xp += src.size();
              }

              lresult.accumulate(kern.at(f),src.at(xp));
            }
            dest.at(x) = lresult.getResult();
          }

          break;

        case lti::Constant:
          for (x=0;x<src.size();++x) {
            for (f = kern.lastIdx(), xx = x+f,lresult.reset();
                 f >= kern.firstIdx(); --f,--xx) {
              if (xx < 0) {
                lresult.accumulate(kern.at(f),src.at(0));
              } else if (xx > src.lastIdx()) {
                lresult.accumulate(kern.at(f),src.at(src.lastIdx()));
              } else {
                lresult.accumulate(kern.at(f),src.at(xx));
              }
            }
            dest.at(x) = lresult.getResult();
          }

          break;
        default:
          throw exception("unknown border type");
      }
    }
  }


  template<class T,class A>
  void lti::convHelper1D<T,A>::applyCol(const matrix<T>& src,
                                        matrix<T>& dest,
                                        const modifier::parameters& param){

    int x,i,f,xx,filterLast,srcSize2;
    int begin,end,col;
    lresult.reset();

    const T *fp,*filterLastp,*filterFirstp;

    // alias for kernel term
    const kernel1D<T>& kern = *kernel;

    // initialize member "norm"
    lresult.setNorm( kern.getNorm() );

    // need kernel as vector
    const vector<T>& filter = kern;

    // last valid index of the filter vector
    filterLast = filter.lastIdx();

    // initialize size of dest (not need to initialization!)
    dest.resize(src.rows(),src.columns(),T(0),false,false);

    // middle block filter (the whole kernel can be applied here!)
    begin = (kern.lastIdx()<0) ? -kern.lastIdx() : 0;
    end   = (kern.lastIdx()>=filter.size()) ?
      src.rows()-kern.lastIdx() :
      src.rows()-filter.size()+1;

    filterFirstp= &filter.at(0);
    --filterFirstp;
    filterLastp = &filter.at(filterLast);

    if (src.rows() >= filter.size()) {
      // ----------------------------------------
      // if the kernel is smaller than the vector

      for (x=begin,i=kern.lastIdx()+x;x<end;++x,++i) {
        // for each column of the matrix do...

        for (col=0;
             col<src.columns();
             ++col) {
          // x: begin in src vector
          // i: place to write local dilation
          // lresult : local result
          // xx: sample on image to be checked
          for (xx=x,lresult.reset(),fp=filterLastp;
               fp!=filterFirstp;--fp,++xx ) {
            lresult.accumulate(*fp,src.at(xx,col));
          }
          dest.at(i,col)=lresult.getResult();
        }
      }

      // -----------------------
      //    border conditions
      // -----------------------

      switch (param.boundaryType) {
        case lti::NoBoundary:
          break;
        case lti::Zero:
          // left border!
          for (x=begin-1,i=kern.lastIdx()+x;
               x>=-kern.lastIdx();
               --x,--i) {
            for (col=0;col<src.columns();++col) {
              for (xx=x+filterLast,lresult.reset(),f=0;
                   ((xx>=0) && (f<filter.size()));
                   ++f,--xx) {
                lresult.accumulate(filter.at(f),src.at(xx,col));
              }
              dest.at(i,col)=lresult.getResult();
            }
          }

          // right border!
          for (x=end,i=kern.lastIdx()+x;
               x<src.rows()-kern.lastIdx();
               ++x,++i) {
            for (col=0;col<src.columns();++col) {
              for (xx=x,lresult.reset(),f=filterLast;
                   ((xx<src.rows()) && (f>=0));
                   --f,++xx) {
                lresult.accumulate(filter.at(f),src.at(xx,col));
              }
              dest.at(i,col)=lresult.getResult();
            }
          }
          break;

        case lti::Mirror:
          // left border!
          for (x=begin-1,i=kern.lastIdx()+x;
               x>=-kern.lastIdx();
               --x,--i) {
            for (col=0;col<src.columns();++col) {
              for (xx=x+filterLast,lresult.reset(),f=0;
                   (f<filter.size());
                   ++f,--xx) {
                if (xx>=0)
                  lresult.accumulate(filter.at(f),src.at(xx,col));
                else
                  lresult.accumulate(filter.at(f),src.at(-xx,col));
              }
              dest.at(i,col)=lresult.getResult();
            }
          }
          // right border!
          srcSize2=2*src.lastRow();
          for (x=end,i=kern.lastIdx()+x;
               x<src.rows()-kern.lastIdx();
               ++x,++i) {
            for (col=0;col<src.columns();++col) {
              for (xx=x,lresult.reset(),f=filterLast;
                   (f>=0);
                   --f,++xx) {
                if (xx<src.rows())
                  lresult.accumulate(filter.at(f),src.at(xx,col));
                else
                  lresult.accumulate(filter.at(f),src.at(srcSize2-xx,col));
              }
              dest.at(i,col)=lresult.getResult();
            }
          }
          break;

        case lti::Periodic:
          // left border!
          for (x=begin-1,i=kern.lastIdx()+x;
               x>=-kern.lastIdx();
               --x,--i) {
            for (col=0;col<src.columns();++col) {
              for (xx=x+filterLast,lresult.reset(),f=0;
                   (f<filter.size());
                   ++f,--xx) {
                if (xx>=0)
                  lresult.accumulate(filter.at(f),src.at(xx,col));
                else
                  lresult.accumulate(filter.at(f),src.at(src.rows()+xx,col));
              }
              dest.at(i,col)=lresult.getResult();
            }
          }

          // right border!
          for (x=end,i=kern.lastIdx()+x;
               x<src.rows()-kern.lastIdx();
               ++x,++i) {
            for (col=0;col<src.columns();++col) {
              for (xx=x,lresult.reset(),f=filterLast;
                   (f>=0);
                   --f,++xx) {
                if (xx<src.rows())
                  lresult.accumulate(filter.at(f),src.at(xx,col));
                else
                  lresult.accumulate(filter.at(f),src.at(xx-src.rows(),col));
              }
              dest.at(i,col)=lresult.getResult();
            }
          }
          break;

        case lti::Constant:
          // left border!
          for (x=begin-1,i=kern.lastIdx()+x;
               x>=-kern.lastIdx();
               --x,--i) {
            for (col=0;col<src.columns();++col) {
              for (xx=x+filterLast,lresult.reset(),f=0;
                   (f<filter.size());
                   ++f,--xx) {
                if (xx>=0)
                  lresult.accumulate(filter.at(f),src.at(xx,col));
                else
                  lresult.accumulate(filter.at(f),src.at(0,col));
              }
              dest.at(i,col)=lresult.getResult();
            }
          }

          // right border!
          for (x=end,i=kern.lastIdx()+x;
               x<src.rows()-kern.lastIdx();
               ++x,++i) {
            for (col=0;col<src.columns();++col) {
              for (xx=x,lresult.reset(),f=filterLast;
                   (f>=0);
                   --f,++xx) {
                if (xx<src.rows())
                  lresult.accumulate(filter.at(f),src.at(xx,col));
                else
                  lresult.accumulate(filter.at(f),src.at(src.lastRow(),col));
              }
              dest.at(i,col)=lresult.getResult();
            }
          }
          break;
        default:
          throw exception("unknown border type");
      }
    } else {
      // ---------------------------------------
      // if the kernel is bigger than the vector

      switch (param.boundaryType) {
        case lti::NoBoundary:
          break;
        case lti::Zero:
          for (x=0;x<src.rows();++x) {
            for (col=0;col<src.columns();++col) {
              for (f = kern.lastIdx(), xx = x+f,lresult.reset();
                   f >= kern.firstIdx(); --f,--xx) {
                if ((xx>=0) && (xx<src.rows())) {
                  lresult.accumulate(kern.at(f),src.at(xx,col));
                }
              }
              dest.at(x,col) = lresult.getResult();
            }
          }
          break;

        case lti::Mirror:
          for (x=0;x<src.rows();++x) {
            for (col=0;col<src.columns();++col) {
              for (f = kern.lastIdx(), xx = x+f,lresult.reset();
                   f >= kern.firstIdx(); --f,--xx) {
                if ((xx>=0) && (xx<src.rows())) {
                  lresult.accumulate(kern.at(f),src.at(xx,col));
                } else {
                  int p,xp;
                  p = xx/src.rows();
                  xp = xx%src.rows();

                  if (xx < 0) {
                    p = -p + 1;
                    if (xp < 0) {
                      xp += src.rows();
                    }
                  }

                  if ((p & 01) != 0) { // even number?
                    xp = src.rows() -1 -xp;
                  }

                  lresult.accumulate(kern.at(f),src.at(xp,col));

                }
              }
              dest.at(x,col) = lresult.getResult();
            }
          }
          break;

        case lti::Periodic:

          for (x=0;x<src.rows();++x) {
            for (col=0;col<src.columns();++col) {
              for (f = kern.lastIdx(), xx = x+f,lresult.reset();
                   f >= kern.firstIdx(); --f,--xx) {
                int xp;

                xp = xx%src.rows();
                if (xp < 0) {
                  xp+=src.rows();
                }
                lresult.accumulate(kern.at(f),src.at(xp,col));
              }
              dest.at(x,col) = lresult.getResult();
            }
          }
          break;

        case lti::Constant:
          for (x=0;x<src.rows();++x) {
            for (col=0;col<src.columns();++col) {
              for (f = kern.lastIdx(), xx = x+f,lresult.reset();
                   f >= kern.firstIdx(); --f,--xx) {
                if (xx < 0) {
                  lresult.accumulate(kern.at(f),src.at(0,col));
                } else if (xx >= src.rows()) {
                  lresult.accumulate(kern.at(f),src.at(src.lastRow(),col));
                } else {
                  lresult.accumulate(kern.at(f),src.at(xx,col));
                }
              }
              dest.at(x,col) = lresult.getResult();
            }
          }

          break;
        default:
          throw exception("unknown border type!");
      }
    }
  }

  template<class T,class A>
  void lti::convHelper1D<T,A>::applyRow(const matrix<T>& src,
                                                         matrix<T>& dest,
                                     const modifier::parameters& param){

    int i;

    dest.resize(src.rows(),src.columns(),T(0),false,false);


    for (i=0;i<src.rows();++i) {
      apply(src.getRow(i),dest.getRow(i),param);
    }

  }

  template<class T,class A>
  A& lti::convHelper1D<T,A>::getAccumulator() {
    return lresult;
  }

  template<class T,class A>
  const A& lti::convHelper1D<T,A>::getAccumulator() const {
    return lresult;
  }

  // ---------------------------------------------
  // 2D FILTERING
  // ---------------------------------------------
  template<class T,class A>
    lti::convHelper2D<T,A>::convHelper2D()
    : kernel(0),sKernel(0),mask(0),tmpKernel(0),tmpSKernel(0) {
  }

  template<class T,class A>
    lti::convHelper2D<T,A>::~convHelper2D() {
    delete tmpKernel;
    tmpKernel=0;
    delete tmpSKernel;
    tmpSKernel=0;
  }

  template<class T,class A>
    bool lti::convHelper2D<T,A>::setKernel(const mathObject* kern) {

    kernel = dynamic_cast<const kernel2D<T>*>(kern);
    sKernel = 0;

    // it is not the expected type, try to cast from know possibilities
    if (isNull(kernel)) {

      sKernel = dynamic_cast<const sepKernel<T>*>(kern);

      if (isNull(sKernel)) {
        // try to get what is it through the name...
        if (*(kern->getTypeName())=='k') {
          // it may be a kernel2D... try it!

          delete tmpKernel;
          tmpKernel = 0;

          // may be a kernel of integers?
          if (notNull(dynamic_cast<const ikernel2D* >(kern))) {
            tmpKernel = new kernel2D<T>;
            tmpKernel->castFrom(*dynamic_cast<const ikernel2D*>(kern));
            if (typeInfo<T>::isFloatingPointType()) {
              tmpKernel->denormalize();
            }
            kernel = tmpKernel;
          } else if (notNull(dynamic_cast<const bkernel2D*>(kern))) {
            tmpKernel = new kernel2D<T>;
            tmpKernel->castFrom(*dynamic_cast<const bkernel2D*>(kern));
            if (typeInfo<T>::isFloatingPointType()) {
              tmpKernel->denormalize();
            }
            kernel = tmpKernel;
          } else if (notNull(dynamic_cast<const fkernel2D*>(kern))) {
            tmpKernel = new kernel2D<T>;

            if (typeInfo<T>::isFloatingPointType()) {
              tmpKernel->castFrom(*dynamic_cast<const fkernel2D*>(kern));
            } else {
              fkernel2D* tmpFloat =
                new fkernel2D(*dynamic_cast<const fkernel2D*>(kern));
              tmpFloat->multiply(static_cast<float>(typeInfo<T>::suggestedNorm()));
              tmpKernel->castFrom(*tmpFloat);
              delete tmpFloat;
            }

            kernel = tmpKernel;
          } else if (notNull(dynamic_cast<const dkernel2D*>(kern))) {
            tmpKernel = new kernel2D<T>;

            if (typeInfo<T>::isFloatingPointType()) {
              tmpKernel->castFrom(*dynamic_cast<const dkernel2D*>(kern));
            } else {
              dkernel2D* tmpFloat =
                new dkernel2D(*dynamic_cast<const dkernel2D*>(kern));
              tmpFloat->multiply(static_cast<double>(typeInfo<T>::suggestedNorm()));
              tmpKernel->castFrom(*tmpFloat);
              delete tmpFloat;
            }

            kernel = tmpKernel;
          }

        } else if (*(kern->getTypeName())=='s') {
          // may be a separable kernel

          delete tmpSKernel;
          tmpSKernel = 0;

          // may be a kernel of integers?
          if (notNull(dynamic_cast<const isepKernel* >(kern))) {
            tmpSKernel = new sepKernel<T>;
            tmpSKernel->castFrom(*dynamic_cast<const isepKernel*>(kern));
            if (typeInfo<T>::isFloatingPointType()) {
              tmpSKernel->denormalize();
            }
            sKernel = tmpSKernel;
          } else if (notNull(dynamic_cast<const bsepKernel*>(kern))) {
            tmpSKernel = new sepKernel<T>;
            tmpSKernel->castFrom(*dynamic_cast<const bsepKernel*>(kern));
            if (typeInfo<T>::isFloatingPointType()) {
              tmpSKernel->denormalize();
            }
            sKernel = tmpSKernel;
          } else if (notNull(dynamic_cast<const fsepKernel*>(kern))) {
            tmpSKernel = new sepKernel<T>;

            if (typeInfo<T>::isFloatingPointType()) {
              tmpSKernel->castFrom(*dynamic_cast<const fsepKernel*>(kern));
            } else {
              fsepKernel* tmpFloat =
                new fsepKernel(*dynamic_cast<const fsepKernel*>(kern));
              tmpFloat->multiply(static_cast<float>(typeInfo<T>::suggestedNorm()));
              tmpSKernel->castFrom(*tmpFloat);
              tmpSKernel->setNorm(typeInfo<T>::suggestedNorm());
              delete tmpFloat;
            }

            sKernel = tmpSKernel;
          } else if (notNull(dynamic_cast<const dsepKernel*>(kern))) {
            tmpSKernel = new sepKernel<T>;

            if (typeInfo<T>::isFloatingPointType()) {
              tmpSKernel->castFrom(*dynamic_cast<const dsepKernel*>(kern));
            } else {
              dsepKernel* tmpFloat =
                new dsepKernel(*dynamic_cast<const dsepKernel*>(kern));
              tmpFloat->multiply(static_cast<double>(typeInfo<T>::suggestedNorm()));
              tmpSKernel->castFrom(*tmpFloat);
              delete tmpFloat;
            }

            sKernel = tmpSKernel;
          }


        }
      }
    }
    return (notNull(kernel) || notNull(sKernel));
  }

  template<class T,class A>
  bool lti::convHelper2D<T,A>::isKernelSeparable() const {
    return notNull(sKernel);
  }

  template<class T,class A>
  void lti::convHelper2D<T,A>::apply(const matrix<T>& src,
                                           matrix<T>& dest,
                                     const modifier::parameters& param) {

    int x,y,i,j,f,g,xx,yy,filterLastRow,filterLastCol;
    int beginRow,endRow,beginCol,endCol;
    lresult.reset();

    // alias for kernel term
    const kernel2D<T>& kern = *kernel;

    if ((kern.columns() < 1) ||
        (kern.rows() < 1)) {
      throw exception("unknown border type!");
    }

    // initialize member "norm"
    lresult.setNorm(kern.getNorm());

    // need kernel as matrix
    const matrix<T>& filter = kern;
    filterLastRow = filter.lastRow();
    filterLastCol = filter.lastColumn();

    // initialize size of dest
    dest.resize(src.rows(),src.columns(),T(0),false,false);

    // middle block filter (the whole kernel can be applied here!)
    beginRow = (kern.lastRow()<0) ? -kern.lastRow() : 0;
    beginCol = (kern.lastColumn()<0) ? -kern.lastColumn() : 0;
    endRow   = (kern.lastRow()>=filter.rows()) ?
                 src.rows()-kern.lastRow() :
                 src.rows()-filter.rows()+1;
    endCol   = (kern.lastColumn()>=filter.columns()) ?
                 src.columns()-kern.lastColumn() :
                 src.columns()-filter.columns()+1;

    if ((src.rows() >= filter.rows()) &&
        (src.columns() >= filter.columns())) {
      // ----------------------------------------
      // if the kernel is smaller than the matrix


      for (y=beginRow,j=kern.lastRow()+y;y<endRow;++y,++j) {
        for (x=beginCol,i=kern.lastColumn()+x;x<endCol;++x,++i) {
          // x,y: begin in src vector
          // i,j: place to write local dilation
          // lresult : local result
          // xx,yy: sample on image to be checked
          for (yy=y,lresult.reset(),g=filterLastRow;g>=0;--g,++yy) {
            for (xx=x,f=filterLastCol;f>=0;--f,++xx ) {
              lresult.accumulate(filter.at(g,f),src.at(yy,xx));
            }
          }
          dest.at(j,i)=lresult.getResult();
        }
      }

      // -----------------------
      //    border conditions
      // -----------------------

      switch (param.boundaryType) {
        case lti::NoBoundary:
          break;
        case lti::Zero: {
          // left border!
          for (y=beginRow,j=kern.lastRow()+y;y<endRow;++y,++j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y,lresult.reset(),g=filterLastRow;g>=0;--g,++yy) {
                for (xx=x+filterLastCol,f=0;
                     ((xx>=0) && (f<filter.columns()));
                     ++f,--xx) {
                  lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // right border!
          for (y=beginRow,j=kern.lastRow()+y;y<endRow;++y,++j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;g>=0;--g,++yy) {
                for (xx=x,f=filterLastCol;
                     ((xx<src.columns()) && (f>=0));
                     --f,++xx) {
                  lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper border!
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=beginCol,i=kern.lastColumn()+x;x<endCol;++x,++i) {
              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   ((yy>=0) && (g<filter.rows()));
                   ++g,--yy) {
                for (xx=x,f=filterLastCol;f>=0;--f,++xx ) {

                  lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom border!
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=beginCol,i=kern.lastColumn()+x;x<endCol;++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   ((yy<src.rows()) && (g>=0));
                   --g,++yy) {
                for (xx=x,f=filterLastCol;f>=0;--f,++xx ) {
                  lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper left
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   ((yy>=0) && (g<filter.rows()));
                   ++g,--yy) {
                for (xx=x+filterLastCol,f=0;
                     ((xx>=0) && (f<filter.columns()));
                     ++f,--xx) {
                  lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper right
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {

              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   ((yy>=0) && (g<filter.rows()));
                   ++g,--yy) {
                for (xx=x,f=filterLastCol;
                     ((xx<src.columns()) && (f>=0));
                     --f,++xx) {
                  lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom left
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   ((yy<src.rows()) && (g>=0));
                   --g,++yy) {
                for (xx=x+filterLastCol,f=0;
                     ((xx>=0) && (f<filter.columns()));
                     ++f,--xx) {
                  lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom right
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   ((yy<src.rows()) && (g>=0));
                   --g,++yy) {
                for (xx=x,f=filterLastCol;
                     ((xx<src.columns()) && (f>=0));
                     --f,++xx) {
                  lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }
        }
        break;

        case lti::Mirror: {
          // left border!
          for (y=beginRow,j=kern.lastRow()+y;y<endRow;++y,++j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y,lresult.reset(),g=filterLastRow;g>=0;--g,++yy) {
                for (xx=x+filterLastCol,f=0;
                     (f<filter.columns());
                     ++f,--xx) {
                  if (xx>=0)
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  else
                    lresult.accumulate(filter.at(g,f),src.at(yy,-xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // right border!
          int srcCol2=2*src.lastColumn();
          for (y=beginRow,j=kern.lastRow()+y;y<endRow;++y,++j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;g>=0;--g,++yy) {
                for (xx=x,f=filterLastCol;
                     (f>=0);
                     --f,++xx) {
                  if (xx<src.columns())
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  else
                    lresult.accumulate(filter.at(g,f),src.at(yy,srcCol2-xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper border!
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=beginCol,i=kern.lastColumn()+x;x<endCol;++x,++i) {
              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   (g<filter.rows());
                   ++g,--yy) {
                for (xx=x,f=filterLastCol;f>=0;--f,++xx ) {
                  if (yy>=0)
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  else
                    lresult.accumulate(filter.at(g,f),src.at(-yy,xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom border!
          int srcRow2=2*src.lastRow();
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=beginCol,i=kern.lastColumn()+x;x<endCol;++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   (g>=0);
                   --g,++yy) {
                for (xx=x,f=filterLastCol;f>=0;--f,++xx ) {
                  if (yy<src.rows())
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  else
                    lresult.accumulate(filter.at(g,f),src.at(srcRow2-yy,xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper left
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   (g<filter.rows());
                   ++g,--yy) {
                for (xx=x+filterLastCol,f=0;
                     (f<filter.columns());
                     ++f,--xx) {
                  lresult.accumulate(filter.at(g,f),
                                     src.at((yy>=0)?yy:-yy,
                                            (xx>=0)?xx:-xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper right
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {

              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   (g<filter.rows());
                   ++g,--yy) {
                for (xx=x,f=filterLastCol;
                     (f>=0);
                     --f,++xx) {
                  lresult.accumulate(filter.at(g,f),
                                     src.at((yy>=0) ? yy : -yy,
                                        (xx<src.columns()) ? xx : srcCol2-xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom left
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   (g>=0);
                   --g,++yy) {
                for (xx=x+filterLastCol,f=0;
                     (f<filter.columns());
                     ++f,--xx) {
                  lresult.accumulate(filter.at(g,f),
                                     src.at((yy<src.rows()) ? yy : srcRow2-yy,
                                            (xx>=0) ? xx : -xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom right
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   (g>=0);
                   --g,++yy) {
                for (xx=x,f=filterLastCol;
                     (f>=0);
                     --f,++xx) {
                  lresult.accumulate(filter.at(g,f),
                                 src.at((yy<src.rows()) ? yy : srcRow2-yy,
                                        (xx<src.columns()) ? xx : srcCol2-xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }
        }
        break;

        case lti::Periodic: {
          // left border!
          for (y=beginRow,j=kern.lastRow()+y;y<endRow;++y,++j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y,lresult.reset(),g=filterLastRow;g>=0;--g,++yy) {
                for (xx=x+filterLastCol,f=0;
                     (f<filter.columns());
                     ++f,--xx) {
                  if (xx>=0)
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  else
                    lresult.accumulate(filter.at(g,f),src.at(yy,src.columns()+xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // right border!
          for (y=beginRow,j=kern.lastRow()+y;y<endRow;++y,++j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;g>=0;--g,++yy) {
                for (xx=x,f=filterLastCol;
                     (f>=0);
                     --f,++xx) {
                  if (xx<src.columns())
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  else
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx-src.columns()));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper border!
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=beginCol,i=kern.lastColumn()+x;x<endCol;++x,++i) {
              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   (g<filter.rows());
                   ++g,--yy) {
                for (xx=x,f=filterLastCol;f>=0;--f,++xx ) {
                  if (yy>=0)
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  else
                    lresult.accumulate(filter.at(g,f),src.at(yy+src.rows(),xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom border!
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=beginCol,i=kern.lastColumn()+x;x<endCol;++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   (g>=0);
                   --g,++yy) {
                for (xx=x,f=filterLastCol;f>=0;--f,++xx ) {
                  if (yy<src.rows())
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  else
                    lresult.accumulate(filter.at(g,f),src.at(yy-src.rows(),xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper left
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   (g<filter.rows());
                   ++g,--yy) {
                for (xx=x+filterLastCol,f=0;
                     (f<filter.columns());
                     ++f,--xx) {
                  lresult.accumulate(filter.at(g,f),
                                     src.at((yy>=0) ? yy : yy+src.rows(),
                                            (xx>=0) ? xx : xx+src.columns()));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper right
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {

              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   (g<filter.rows());
                   ++g,--yy) {
                for (xx=x,f=filterLastCol;
                     (f>=0);
                     --f,++xx) {
                  lresult.accumulate(filter.at(g,f),
                               src.at((yy>=0)?yy:yy+src.rows(),
                                      (xx<src.columns())?xx:xx-src.columns()));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom left
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   (g>=0);
                   --g,++yy) {
                for (xx=x+filterLastCol,f=0;
                     (f<filter.columns());
                     ++f,--xx) {
                  lresult.accumulate(filter.at(g,f),
                             src.at((yy<src.rows())?yy:yy-src.rows(),
                                    (xx>=0)?xx:xx+src.columns()));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom right
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   (g>=0);
                   --g,++yy) {
                for (xx=x,f=filterLastCol;
                     (f>=0);
                     --f,++xx) {
                  lresult.accumulate(filter.at(g,f),
                             src.at((yy<src.rows())?yy:yy-src.rows(),
                                    (xx<src.columns())?xx:xx-src.columns()));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }
        }
        break;

        case lti::Constant: {
          // left border!
          for (y=beginRow,j=kern.lastRow()+y;y<endRow;++y,++j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y,lresult.reset(),g=filterLastRow;g>=0;--g,++yy) {
                for (xx=x+filterLastCol,f=0;
                     (f<filter.columns());
                     ++f,--xx) {
                  if (xx>=0)
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  else
                    lresult.accumulate(filter.at(g,f),src.at(yy,0));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // right border!
          for (y=beginRow,j=kern.lastRow()+y;y<endRow;++y,++j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;g>=0;--g,++yy) {
                for (xx=x,f=filterLastCol;
                     (f>=0);
                     --f,++xx) {
                  if (xx<src.columns())
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  else
                    lresult.accumulate(filter.at(g,f),src.at(yy,src.lastColumn()));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper border!
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=beginCol,i=kern.lastColumn()+x;x<endCol;++x,++i) {
              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   (g<filter.rows());
                   ++g,--yy) {
                for (xx=x,f=filterLastCol;f>=0;--f,++xx ) {
                  if (yy>=0)
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  else
                    lresult.accumulate(filter.at(g,f),src.at(0,xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom border!
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=beginCol,i=kern.lastColumn()+x;x<endCol;++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   (g>=0);
                   --g,++yy) {
                for (xx=x,f=filterLastCol;f>=0;--f,++xx ) {
                  if (yy<src.rows())
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  else
                    lresult.accumulate(filter.at(g,f),src.at(src.lastRow(),xx));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper left
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   (g<filter.rows());
                   ++g,--yy) {
                for (xx=x+filterLastCol,f=0;
                     (f<filter.columns());
                     ++f,--xx) {
                  lresult.accumulate(filter.at(g,f),
                             src.at((yy>=0) ? yy : 0,
                                    (xx>=0) ? xx : 0));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper right
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {

              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   (g<filter.rows());
                   ++g,--yy) {
                for (xx=x,f=filterLastCol;
                     (f>=0);
                     --f,++xx) {
                  lresult.accumulate(filter.at(g,f),
                             src.at((yy>=0)?yy:0,
                                    (xx<src.columns())?xx:src.lastColumn()));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom left
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   (g>=0);
                   --g,++yy) {
                for (xx=x+filterLastCol,f=0;
                     (f<filter.columns());
                     ++f,--xx) {
                  lresult.accumulate(filter.at(g,f),
                             src.at((yy<src.rows()) ? yy : src.lastRow(),
                                    (xx>=0) ? xx : 0));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom right
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   (g>=0);
                   --g,++yy) {
                for (xx=x,f=filterLastCol;
                     (f>=0);
                     --f,++xx) {
                  lresult.accumulate(filter.at(g,f),
                            src.at((yy<src.rows()) ? yy : src.lastRow(),
                                   (xx<src.columns()) ? xx : src.lastColumn()));
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }
        }
        break;

        default:
          throw exception("unknown border type!");
      }
    } else {
      // ---------------------------------------
      // if the kernel is bigger than the vector

      switch (param.boundaryType) {
        case lti::NoBoundary:
          break;
        case lti::Zero:
          for (y=0;y<src.rows();++y) {
            for (x=0;x<src.columns();++x) {
              for (g = kern.lastRow(), yy = y+g, lresult.reset();
                   g >=kern.firstRow();--g,--yy) {
                for (f = kern.lastColumn(), xx = x+f;
                     f >= kern.firstColumn(); --f,--xx) {
                  if ((xx>=0) && (xx<src.columns()) &&
                      (yy>=0) && (yy<src.rows())) {
                    lresult.accumulate(kern.at(g,f),src.at(yy,xx));
                  }
                }
              }
              dest.at(y,x) = lresult.getResult();
            }
          }
          break;

        case lti::Mirror:
          for (y=0;y<src.rows();++y) {
            for (x=0;x<src.columns();++x) {

              for (g = kern.lastRow(), yy = y+g, lresult.reset();
                   g >=kern.firstRow();--g,--yy) {
                for (f = kern.lastColumn(), xx = x+f;
                     f >= kern.firstColumn(); --f,--xx) {

                  if ((xx>=0) && (xx<src.columns()) &&
                      (yy>=0) && (yy<src.rows())) {
                    lresult.accumulate(kern.at(g,f),src.at(yy,xx));
                  } else {
                    int px,py,xp,yp;

                    px = xx/src.columns();
                    py = yy/src.rows();

                    xp = xx % src.columns();
                    yp = yy % src.rows();

                    if (xx < 0) {
                      px = -px + 1;
                      if (xp < 0) {
                        xp += src.columns();
                      }
                    }

                    if (yy < 0) {
                      py = -py + 1;
                      if (yp < 0) {
                        yp += src.rows();
                      }
                    }

                    xp = ((px & 01) == 0) ? xp % src.columns() :
                                            src.lastColumn()-(xp%src.columns());
                    yp = ((py & 01) == 0) ? yp % src.rows() :
                                            src.lastRow()-(yp%src.rows());
                    lresult.accumulate(kern.at(g,f),src.at(yp,xp));
                  }
                }
              }
              dest.at(y,x) = lresult.getResult();
            }
          }
          break;

        case lti::Periodic:

          for (y=0;y<src.rows();++y) {
            for (x=0;x<src.columns();++x) {
              for (g = kern.lastRow(), yy = y+g, lresult.reset();
                   g >=kern.firstRow();--g,--yy) {
                for (f = kern.lastColumn(), xx = x+f;
                     f >= kern.firstColumn(); --f,--xx) {
                  lresult.accumulate(kern.at(g,f),
                             src.at(yy%src.rows(),xx%src.columns()));
                }
              }
              dest.at(y,x) = lresult.getResult();
            }
          }

          break;

        case lti::Constant:
          for (y=0;y<src.rows();++y) {
            for (x=0;x<src.columns();++x) {

              for (g = kern.lastRow(), yy = y+g, lresult.reset();
                   g >=kern.firstRow();--g,--yy) {
                for (f = kern.lastColumn(), xx = x+f;
                     f >= kern.firstColumn(); --f,--xx) {

                  int xp,yp;
                  yp = (yy<0) ? 0 : (yy>=src.rows()) ? src.lastRow() : yy;
                  xp = (xx<0) ? 0 : (xx>=src.rows()) ? src.lastRow() : xx;

                  lresult.accumulate(kern.at(g,f),src.at(yp,xp));
                }
              }
              dest.at(y,x) = lresult.getResult();
            }
          }

          break;
        default:
          throw exception("unknown border type!");
      }
    }
  }

  template<class T,class A>
  void lti::convHelper2D<T,A>::applySep(const matrix<T>& src,
                                               matrix<T>& dest,
                                         const modifier::parameters& param){


    // alias for kernel term
    const sepKernel<T>& kern = *sKernel;

    int i;
    matrix<T> tmpA,tmpB;

    // initialize member "norm"
    lti::convHelper1D<T,A> conv1D;

    for (i=0;i<kern.getNumberOfPairs();++i) {
      // rows filter
      if ((kern.getRowFilter(i).size() == 1) &&
          (kern.getRowFilter(i).getOffset() == 0)) {
        tmpA.multiply(src,kern.getRowFilter(i).at(0));
      } else {
        conv1D.setKernel(&kern.getRowFilter(i));
        conv1D.applyRow(src,tmpA,param);
      }

      // TODO: sometimes is better to transpose the matrix and then
      //       use applyRow again!  This should be tested!

      // columns filter
      if ((kern.getColFilter(i).size() == 1) &&
          (kern.getColFilter(i).getOffset() == 0)) {
        tmpB.multiply(tmpA,kern.getColFilter(i).at(0));
      } else {
        conv1D.setKernel(&kern.getColFilter(i));
        conv1D.applyCol(tmpA,tmpB,param);
      }

      if (i==0) {
        // the first time just transfer the result to dest
        tmpB.detach(dest);
      } else {
        // the rest separable parts should be accumulated in the result!
        dest.add(tmpB);
      }
    }
  }

  template<class T,class A>
  A& lti::convHelper2D<T,A>::getAccumulator() {
    return lresult;
  }

  template<class T,class A>
  const A& lti::convHelper2D<T,A>::getAccumulator() const {
    return lresult;
  }

  template<class T,class A>
  void lti::convHelper2D<T,A>::applyMask(const matrix<T>& src,
                                           matrix<T>& dest,
                                     const modifier::parameters& param) {
    // no correlation possible without a matching mask
    if (isNull(mask) || (mask->size() != kernel->size())) {
      throw functor::invalidParametersException();
    }

    int x,y,i,j,xx,yy,f,g,filterLastRow,filterLastCol;
    int beginRow,endRow,beginCol,endCol;
    lresult.reset();

    // alias for kernel term
    const kernel2D<T>& kern = *kernel;

    if ((kern.columns() < 1) ||
        (kern.rows() < 1)) {
      throw exception("unknown border type!");
    }

    // initialize member "norm"
    lresult.setNorm(kern.getNorm());

    // need kernel as matrix
    const matrix<T>& filter = kern;
    filterLastRow = filter.lastRow();
    filterLastCol = filter.lastColumn();

    // initialize size of dest
    dest.resize(src.rows(),src.columns(),T(0),false,true/*false*/);

    // middle block filter (the whole kernel can be applied here!)
    beginRow = (kern.lastRow()<0) ? -kern.lastRow() : 0;
    beginCol = (kern.lastColumn()<0) ? -kern.lastColumn() : 0;
    endRow   = (kern.lastRow()>=filter.rows()) ?
                 src.rows()-kern.lastRow() :
                 src.rows()-filter.rows()+1;
    endCol   = (kern.lastColumn()>=filter.columns()) ?
                 src.columns()-kern.lastColumn() :
                 src.columns()-filter.columns()+1;

    lti::pointList::const_iterator p;

    if ((src.rows() >= filter.rows()) &&
        (src.columns() >= filter.columns())) {
      // ----------------------------------------
      // if the kernel is smaller than the matrix

      lti::areaPoints shape;
      shape.getFromMask(*mask);

      for (y=beginRow,j=kern.lastRow()+y;y<endRow;++y,++j) {
        for (x=beginCol,i=kern.lastColumn()+x;x<endCol;++x,++i) {
          // x,y: begin in src vector
          // i,j: place to write local dilation
          // lresult : local result
          // xx,yy: sample on image to be checked

          lresult.reset();
          for (p = shape.begin(); p != shape.end(); p++) {
            lresult.accumulate(filter.at(*p),
              src.at(y+filterLastRow - p->y,x+filterLastCol - p->x));
          }
          dest.at(j,i)=lresult.getResult();
        }
      }

      // -----------------------
      //    border conditions
      // -----------------------

      switch (param.boundaryType) {
        case lti::NoBoundary:
          break;
        case lti::Zero: {
          // left border!
          for (y=beginRow,j=kern.lastRow()+y;y<endRow;++y,++j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y,lresult.reset(),g=filterLastRow;g>=0;--g,++yy) {
                for (xx=x+filterLastCol,f=0;
                     ((xx>=0) && (f<filter.columns()));
                     ++f,--xx) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // right border!
          for (y=beginRow,j=kern.lastRow()+y;y<endRow;++y,++j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;g>=0;--g,++yy) {
                for (xx=x,f=filterLastCol;
                     ((xx<src.columns()) && (f>=0));
                     --f,++xx) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper border!
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=beginCol,i=kern.lastColumn()+x;x<endCol;++x,++i) {
              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   ((yy>=0) && (g<filter.rows()));
                   ++g,--yy) {
                for (xx=x,f=filterLastCol;f>=0;--f,++xx ) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom border!
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=beginCol,i=kern.lastColumn()+x;x<endCol;++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   ((yy<src.rows()) && (g>=0));
                   --g,++yy) {
                for (xx=x,f=filterLastCol;f>=0;--f,++xx ) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper left
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   ((yy>=0) && (g<filter.rows()));
                   ++g,--yy) {
                for (xx=x+filterLastCol,f=0;
                     ((xx>=0) && (f<filter.columns()));
                     ++f,--xx) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper right
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {

              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   ((yy>=0) && (g<filter.rows()));
                   ++g,--yy) {
                for (xx=x,f=filterLastCol;
                     ((xx<src.columns()) && (f>=0));
                     --f,++xx) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom left
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   ((yy<src.rows()) && (g>=0));
                   --g,++yy) {
                for (xx=x+filterLastCol,f=0;
                     ((xx>=0) && (f<filter.columns()));
                     ++f,--xx) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom right
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   ((yy<src.rows()) && (g>=0));
                   --g,++yy) {
                for (xx=x,f=filterLastCol;
                     ((xx<src.columns()) && (f>=0));
                     --f,++xx) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }
        }
        break;

        case lti::Mirror: {
          // left border!
          for (y=beginRow,j=kern.lastRow()+y;y<endRow;++y,++j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y,lresult.reset(),g=filterLastRow;g>=0;--g,++yy) {
                for (xx=x+filterLastCol,f=0;
                     (f<filter.columns());
                     ++f,--xx) {
                  if (mask->at(g,f) != 0) {
                    if (xx>=0)
                      lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                    else
                      lresult.accumulate(filter.at(g,f),src.at(yy,-xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // right border!
          int srcCol2=2*src.lastColumn();
          for (y=beginRow,j=kern.lastRow()+y;y<endRow;++y,++j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;g>=0;--g,++yy) {
                for (xx=x,f=filterLastCol;
                     (f>=0);
                     --f,++xx) {
                  if (mask->at(g,f) != 0) {
                    if (xx<src.columns())
                      lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                    else
                      lresult.accumulate(filter.at(g,f),src.at(yy,srcCol2-xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper border!
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=beginCol,i=kern.lastColumn()+x;x<endCol;++x,++i) {
              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   (g<filter.rows());
                   ++g,--yy) {
                for (xx=x,f=filterLastCol;f>=0;--f,++xx ) {
                  if (mask->at(g,f) != 0) {
                    if (yy>=0)
                      lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                    else
                      lresult.accumulate(filter.at(g,f),src.at(-yy,xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom border!
          int srcRow2=2*src.lastRow();
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=beginCol,i=kern.lastColumn()+x;x<endCol;++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   (g>=0);
                   --g,++yy) {
                for (xx=x,f=filterLastCol;f>=0;--f,++xx ) {
                  if (mask->at(g,f) != 0) {
                    if (yy<src.rows())
                      lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                    else
                      lresult.accumulate(filter.at(g,f),src.at(srcRow2-yy,xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper left
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   (g<filter.rows());
                   ++g,--yy) {
                for (xx=x+filterLastCol,f=0;
                     (f<filter.columns());
                     ++f,--xx) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),
                                       src.at((yy>=0)?yy:-yy,
                                              (xx>=0)?xx:-xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper right
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {

              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   (g<filter.rows());
                   ++g,--yy) {
                for (xx=x,f=filterLastCol;
                     (f>=0);
                     --f,++xx) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),
                                       src.at((yy>=0) ? yy : -yy,
                                          (xx<src.columns()) ? xx : srcCol2-xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom left
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   (g>=0);
                   --g,++yy) {
                for (xx=x+filterLastCol,f=0;
                     (f<filter.columns());
                     ++f,--xx) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),
                                       src.at((yy<src.rows()) ? yy : srcRow2-yy,
                                              (xx>=0) ? xx : -xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom right
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   (g>=0);
                   --g,++yy) {
                for (xx=x,f=filterLastCol;
                     (f>=0);
                     --f,++xx) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),
                                   src.at((yy<src.rows()) ? yy : srcRow2-yy,
                                          (xx<src.columns()) ? xx : srcCol2-xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }
        }
        break;

        case lti::Periodic: {
          // left border!
          for (y=beginRow,j=kern.lastRow()+y;y<endRow;++y,++j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y,lresult.reset(),g=filterLastRow;g>=0;--g,++yy) {
                for (xx=x+filterLastCol,f=0;
                     (f<filter.columns());
                     ++f,--xx) {
                  if (mask->at(g,f) != 0) {
                    if (xx>=0)
                      lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                    else
                      lresult.accumulate(filter.at(g,f),src.at(yy,src.columns()+xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // right border!
          for (y=beginRow,j=kern.lastRow()+y;y<endRow;++y,++j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;g>=0;--g,++yy) {
                for (xx=x,f=filterLastCol;
                     (f>=0);
                     --f,++xx) {
                  if (mask->at(g,f) != 0) {
                    if (xx<src.columns())
                      lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                    else
                      lresult.accumulate(filter.at(g,f),src.at(yy,xx-src.columns()));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper border!
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=beginCol,i=kern.lastColumn()+x;x<endCol;++x,++i) {
              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   (g<filter.rows());
                   ++g,--yy) {
                for (xx=x,f=filterLastCol;f>=0;--f,++xx ) {
                  if (mask->at(g,f) != 0) {
                    if (yy>=0)
                      lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                    else
                      lresult.accumulate(filter.at(g,f),src.at(yy+src.rows(),xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom border!
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=beginCol,i=kern.lastColumn()+x;x<endCol;++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   (g>=0);
                   --g,++yy) {
                for (xx=x,f=filterLastCol;f>=0;--f,++xx ) {
                  if (mask->at(g,f) != 0) {
                    if (yy<src.rows())
                      lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                    else
                      lresult.accumulate(filter.at(g,f),src.at(yy-src.rows(),xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper left
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   (g<filter.rows());
                   ++g,--yy) {
                for (xx=x+filterLastCol,f=0;
                     (f<filter.columns());
                     ++f,--xx) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),
                                       src.at((yy>=0) ? yy : yy+src.rows(),
                                              (xx>=0) ? xx : xx+src.columns()));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper right
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {

              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   (g<filter.rows());
                   ++g,--yy) {
                for (xx=x,f=filterLastCol;
                     (f>=0);
                     --f,++xx) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),
                                 src.at((yy>=0)?yy:yy+src.rows(),
                                        (xx<src.columns())?xx:xx-src.columns()));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom left
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   (g>=0);
                   --g,++yy) {
                for (xx=x+filterLastCol,f=0;
                     (f<filter.columns());
                     ++f,--xx) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),
                               src.at((yy<src.rows())?yy:yy-src.rows(),
                                      (xx>=0)?xx:xx+src.columns()));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom right
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   (g>=0);
                   --g,++yy) {
                for (xx=x,f=filterLastCol;
                     (f>=0);
                     --f,++xx) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),
                               src.at((yy<src.rows())?yy:yy-src.rows(),
                                      (xx<src.columns())?xx:xx-src.columns()));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }
        }
        break;

        case lti::Constant: {
          // left border!
          for (y=beginRow,j=kern.lastRow()+y;y<endRow;++y,++j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y,lresult.reset(),g=filterLastRow;g>=0;--g,++yy) {
                for (xx=x+filterLastCol,f=0;
                     (f<filter.columns());
                     ++f,--xx) {
                  if (mask->at(g,f) != 0) {
                    if (xx>=0)
                      lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                    else
                      lresult.accumulate(filter.at(g,f),src.at(yy,0));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // right border!
          for (y=beginRow,j=kern.lastRow()+y;y<endRow;++y,++j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;g>=0;--g,++yy) {
                for (xx=x,f=filterLastCol;
                     (f>=0);
                     --f,++xx) {
                  if (mask->at(g,f) != 0) {
                    if (xx<src.columns())
                      lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                    else
                      lresult.accumulate(filter.at(g,f),src.at(yy,src.lastColumn()));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper border!
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=beginCol,i=kern.lastColumn()+x;x<endCol;++x,++i) {
              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   (g<filter.rows());
                   ++g,--yy) {
                for (xx=x,f=filterLastCol;f>=0;--f,++xx ) {
                  if (mask->at(g,f) != 0) {
                    if (yy>=0)
                      lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                    else
                      lresult.accumulate(filter.at(g,f),src.at(0,xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom border!
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=beginCol,i=kern.lastColumn()+x;x<endCol;++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   (g>=0);
                   --g,++yy) {
                for (xx=x,f=filterLastCol;f>=0;--f,++xx ) {
                  if (mask->at(g,f) != 0) {
                    if (yy<src.rows())
                      lresult.accumulate(filter.at(g,f),src.at(yy,xx));
                    else
                      lresult.accumulate(filter.at(g,f),src.at(src.lastRow(),xx));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper left
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   (g<filter.rows());
                   ++g,--yy) {
                for (xx=x+filterLastCol,f=0;
                     (f<filter.columns());
                     ++f,--xx) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),
                               src.at((yy>=0) ? yy : 0,
                                      (xx>=0) ? xx : 0));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // upper right
          for (y=beginRow-1,j=kern.lastRow()+y;
               y>=-kern.lastRow();
               --y,--j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {

              for (yy=y+filterLastRow,lresult.reset(),g=0;
                   (g<filter.rows());
                   ++g,--yy) {
                for (xx=x,f=filterLastCol;
                     (f>=0);
                     --f,++xx) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),
                               src.at((yy>=0)?yy:0,
                                      (xx<src.columns())?xx:src.lastColumn()));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom left
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=beginCol-1,i=kern.lastColumn()+x;
                 x>=-kern.lastColumn();
                 --x,--i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   (g>=0);
                   --g,++yy) {
                for (xx=x+filterLastCol,f=0;
                     (f<filter.columns());
                     ++f,--xx) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),
                               src.at((yy<src.rows()) ? yy : src.lastRow(),
                                      (xx>=0) ? xx : 0));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }

          // bottom right
          for (y=endRow,j=kern.lastRow()+y;
               y<src.rows()-kern.lastRow();
               ++y,++j) {
            for (x=endCol,i=kern.lastColumn()+x;
                 x<src.columns()-kern.lastColumn();
                 ++x,++i) {
              for (yy=y,lresult.reset(),g=filterLastRow;
                   (g>=0);
                   --g,++yy) {
                for (xx=x,f=filterLastCol;
                     (f>=0);
                     --f,++xx) {
                  if (mask->at(g,f) != 0) {
                    lresult.accumulate(filter.at(g,f),
                              src.at((yy<src.rows()) ? yy : src.lastRow(),
                                     (xx<src.columns()) ? xx : src.lastColumn()));
                  }
                }
              }
              dest.at(j,i)=lresult.getResult();
            }
          }
        }
        break;

        default:
          throw exception("unknown border type!");
      }
    } else {
      // ---------------------------------------
      // if the kernel is bigger than the matrix

      int kernFirstRow = kern.firstRow();
      int kernFirstCol = kern.firstColumn();

      switch (param.boundaryType) {
        case lti::NoBoundary:
          break;
        case lti::Zero:
          for (y=0;y<src.rows();++y) {
            for (x=0;x<src.columns();++x) {
              for (g = kern.lastRow(), yy = y+g, lresult.reset();
                   g >=kern.firstRow();--g,--yy) {
                for (f = kern.lastColumn(), xx = x+f;
                     f >= kern.firstColumn(); --f,--xx) {
                  if ((xx>=0) && (xx<src.columns()) &&
                      (yy>=0) && (yy<src.rows()) &&
                      (mask->at(g - kernFirstRow, f - kernFirstCol) != 0)) {
                    lresult.accumulate(kern.at(g,f),src.at(yy,xx));
                  }
                }
              }
              dest.at(y,x) = lresult.getResult();
            }
          }
          break;

        case lti::Mirror:
          for (y=0;y<src.rows();++y) {
            for (x=0;x<src.columns();++x) {

              for (g = kern.lastRow(), yy = y+g, lresult.reset();
                   g >=kern.firstRow();--g,--yy) {
                for (f = kern.lastColumn(), xx = x+f;
                     f >= kern.firstColumn(); --f,--xx) {
                  if (mask->at(g - kernFirstRow, f - kernFirstCol) != 0) {
                    continue;
                  }
                  if ((xx>=0) && (xx<src.columns()) &&
                      (yy>=0) && (yy<src.rows())) {
                    lresult.accumulate(kern.at(g,f),src.at(yy,xx));
                  } else {
                    int px,py,xp,yp;

                    px = xx/src.columns();
                    py = yy/src.rows();

                    xp = xx % src.columns();
                    yp = yy % src.rows();

                    if (xx < 0) {
                      px = -px + 1;
                      if (xp < 0) {
                        xp += src.columns();
                      }
                    }

                    if (yy < 0) {
                      py = -py + 1;
                      if (yp < 0) {
                        yp += src.rows();
                      }
                    }

                    xp = ((px & 01) == 0) ? xp % src.columns() :
                                            src.lastColumn()-(xp%src.columns());
                    yp = ((py & 01) == 0) ? yp % src.rows() :
                                            src.lastRow()-(yp%src.rows());
                    lresult.accumulate(kern.at(g,f),src.at(yp,xp));
                  }
                }
              }
              dest.at(y,x) = lresult.getResult();
            }
          }
          break;

        case lti::Periodic:

          for (y=0;y<src.rows();++y) {
            for (x=0;x<src.columns();++x) {
              for (g = kern.lastRow(), yy = y+g, lresult.reset();
                   g >=kern.firstRow();--g,--yy) {
                for (f = kern.lastColumn(), xx = x+f;
                     f >= kern.firstColumn(); --f,--xx) {
                  if (mask->at(g - kernFirstRow, f - kernFirstCol) != 0) {
                    lresult.accumulate(kern.at(g,f),
                               src.at(yy%src.rows(),xx%src.columns()));
                  }
                }
              }
              dest.at(y,x) = lresult.getResult();
            }
          }

          break;

        case lti::Constant:
          for (y=0;y<src.rows();++y) {
            for (x=0;x<src.columns();++x) {

              for (g = kern.lastRow(), yy = y+g, lresult.reset();
                   g >=kern.firstRow();--g,--yy) {
                for (f = kern.lastColumn(), xx = x+f;
                     f >= kern.firstColumn(); --f,--xx) {
                  if (mask->at(g - kernFirstRow, f - kernFirstCol) != 0) {
                    int xp,yp;
                    yp = (yy<0) ? 0 : (yy>=src.rows()) ? src.lastRow() : yy;
                    xp = (xx<0) ? 0 : (xx>=src.rows()) ? src.lastRow() : xx;

                    lresult.accumulate(kern.at(g,f),src.at(yp,xp));
                  }
                }
              }
              dest.at(y,x) = lresult.getResult();
            }
          }

          break;
        default:
          throw exception("unknown border type!");
      }
    }
  }

  template<class T,class A>
  void lti::convHelper2D<T,A>::setMask(const channel8& mask) {
    this->mask = &mask;
  }

  template<class T,class A>
  const channel8* lti::convHelper2D<T,A>::getMask() const {
    return mask;
  }

}

