/*
 * Copyright (C) 1998, 1999, 2000, 2001
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 *
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.
 */

/*----------------------------------------------------------------
 * project ....: LTI Digital Image/Signal Processing Library
 * file .......: ltiGeometricFeatureGroup.h
 * authors ....: Claudia Goenner
 * organization: LTI, RWTH Aachen
 * creation ...: 22.07.2003
 * revisions ..:
 */

#include "ltiGeometricFeatureGroup.h"

namespace lti {

  geometricFeatureGroup0::geometricFeatureGroup0()
    : minX(), maxX(), minY(), maxY(), area(0), cog() 
  {}

  geometricFeatureGroup0::geometricFeatureGroup0
  (const point& initMinX, const point& initMaxX,
   const point& initMinY, const point& initMaxY,
   const int initArea, const tpoint<float>& initCog)
    : minX(initMinX), maxX(initMaxX),
      minY(initMinY), maxY(initMaxY), area(initArea), cog(initCog)
  {}

  geometricFeatureGroup0::geometricFeatureGroup0(const geometricFeatureGroup0& other)
  {
    copy(other);
  }

  geometricFeatureGroup0& 
  geometricFeatureGroup0::copy(const geometricFeatureGroup0& other)
  {
    area = other.area;
    minX = other.minX;
    maxX = other.maxX;
    minY = other.minY;
    maxY = other.maxY;
    cog.copy(other.cog);
    
    return *this;
  }

  geometricFeatureGroup0& 
  geometricFeatureGroup0::operator=(const geometricFeatureGroup0& other) {
    return (copy(other));
  }

  bool geometricFeatureGroup0::addObjectPoint(int x, int y) {

    if ( x < minX.x ) {
      minX.x = x;  minX.y = y;
    }
    if ( x > maxX.x ) {
      maxX.x = x;  maxX.y = y;
    } 
    if ( y < minY.y ) {
      minY.x = x;  minY.y = y;
    }
    if ( y > maxY.y ) {
      maxY.x = x;  maxY.y = y;
    }  
 
    area++;
    cog.x += float(x);
    cog.y += float(y);
    return true;
  }

  bool geometricFeatureGroup0::addObjectPoint(point p) {
    return addObjectPoint(p.x, p.y);
  }

  bool geometricFeatureGroup0::completeComputation() {
    
    if (area > 0) {
      cog /= float(area);
      return true;
      
    } else
      return false;
  }

  bool geometricFeatureGroup0::operator==(const geometricFeatureGroup0& other) const
  {
    return ( minX == other.minX && maxX == other.maxX
	     && minY == other.minY && maxY == other.maxY
	     && area == other.area && cog == other.cog );
  }

  bool geometricFeatureGroup0::operator!=(const geometricFeatureGroup0& other) const
  {
    return ( minX != other.minX && maxX != other.maxX
	     && minY != other.minY && maxY != other.maxY
	     && area != other.area && cog != other.cog );
  }

  bool geometricFeatureGroup0::isClose(const geometricFeatureGroup0& other, const point& dist) const {
    
    const point br1 ( other.maxX.x + dist.x,
		      other.maxY.y + dist.y );
    const point ul1 ( other.minX.x - dist.x,
		      other.minY.y - dist.y );
    
    if ( maxX.x >= ul1.x ) {
      if ( maxX.x <= br1.x ) {
	if ( maxY.y >= ul1.y ) {
	  if ( maxY.y <= br1.y ) {
	    return true;
	    
	  } else if ( minY.y <= br1.y ) {
	    return true;

	  } else {
	    return false;
	  }

	} else {
	  return false;
	}

      } else if ( minX.x <= br1.x ) {
	if ( minY.y >= ul1.y ) {
	  if ( minY.y <= br1.y ) {
	    return true;

	  } else {
	    return false;
	  }

	} else if ( maxY.y >= ul1.y ) {
	  if ( minY.y <= br1.y ) {
	    return true;

	  } else {
	    return false;
	  }

	} else {
	  return false;
	}
      }
    }
    
    return false;
  }

  bool  read(ioHandler& handler, geometricFeatureGroup0& p, const bool complete)
  {
    bool b;

    if (complete) {
      handler.readBegin();
    }

    b = read(handler, p.minX);

    handler.readDataSeparator();
    b = read(handler, p.maxX);

    handler.readDataSeparator();
    b = read(handler, p.minY);

    handler.readDataSeparator();
    b = read(handler, p.maxY);

    handler.readDataSeparator();
    b = read(handler, p.area);

    handler.readDataSeparator();
    b = read(handler, p.cog);

    if (complete) {
      b = handler.readEnd();
    }
    return b;
  }

  bool write(ioHandler& handler, const geometricFeatureGroup0& p, const bool complete)
  {
    bool b;
    
    if (complete) {
      handler.writeBegin();
    }
    
    b = write(handler, p.minX);

    handler.writeDataSeparator();
    b = write(handler, p.maxX);

    handler.writeDataSeparator();
    b = write(handler, p.minY);

    handler.writeDataSeparator();
    b = write(handler, p.maxY);

    handler.writeDataSeparator();
    b = write(handler, p.area);
    handler.writeDataSeparator();
    b = write(handler, p.cog);

    if (complete) {
      b = handler.writeEnd();
    }

    return b;
  }

}//namespace lti
