/*
 * Copyright (C) 1998, 1999, 2000, 2001
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 *
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.
 */


/* -------------------------------------------------------------------
 * project ....: LTI-Lib: Image Processing and Computer Vision Library
 * file .......: ltiHarrisCorners.cpp
 * authors ....: Pablo Alvarado
 * organization: LTI, RWTH Aachen
 * creation ...: 31.10.2002
 * revisions ..: $Id: ltiHarrisCorners.cpp,v 1.10 2003/11/17 16:04:43 alvarado Exp $
 */

#include "ltiMacroSymbols.h"

#ifdef _LTI_MSC_6
#pragma warning( disable : 4786 )
#endif

#include "ltiHarrisCorners.h"
#include "ltiConvolution.h"
#include "ltiGradientKernels.h"
#include "ltiLinearKernels.h"
#include "ltiGaussKernels.h"
#include "ltiLocalMaxima.h"

namespace lti {
  // --------------------------------------------------
  // harrisCorners::parameters
  // --------------------------------------------------

  // default constructor
  harrisCorners::parameters::parameters()
    : cornerDetector::parameters() {
    variance = float(-1);
    kernelSize = int(7);
    maximumCorners = int(300);
    scale = float(0.04f);
    gradientMode = HarrisGradient;
  }

  // copy constructor
  harrisCorners::parameters::parameters(const parameters& other)
    : cornerDetector::parameters()  {
    copy(other);
  }

  // destructor
  harrisCorners::parameters::~parameters() {
  }

  // get type name
  const char* harrisCorners::parameters::getTypeName() const {
    return "harrisCorners::parameters";
  }

  // copy member

  harrisCorners::parameters&
    harrisCorners::parameters::copy(const parameters& other) {
# ifndef _LTI_MSC_VER
    // MS Visual C++ 6 is not able to compile this...
    cornerDetector::parameters::copy(other);
# else
    // ...so we have to use this workaround.
    // Conditional on that, copy may not be virtual.
    cornerDetector::parameters& (cornerDetector::parameters::* p_copy)
      (const cornerDetector::parameters&) =
      cornerDetector::parameters::copy;
    (this->*p_copy)(other);
# endif


    variance = other.variance;
    kernelSize = other.kernelSize;
    maximumCorners = other.maximumCorners;
    scale = other.scale;
    localMaximaParameters.copy(other.localMaximaParameters);
    gradientMode = other.gradientMode;

    return *this;
  }

  // alias for copy member
  harrisCorners::parameters&
    harrisCorners::parameters::operator=(const parameters& other) {
    return copy(other);
  }

  // clone member
  functor::parameters* harrisCorners::parameters::clone() const {
    return new parameters(*this);
  }

  /*
   * write the parameters in the given ioHandler
   * @param handler the ioHandler to be used
   * @param complete if true (the default) the enclosing begin/end will
   *        be also written, otherwise only the data block will be written.
   * @return true if write was succeful
   */
# ifndef _LTI_MSC_VER
  bool harrisCorners::parameters::write(ioHandler& handler,
                                         const bool& complete) const
# else
  bool harrisCorners::parameters::writeMS(ioHandler& handler,
                                           const bool& complete) const
# endif
  {
    bool b = true;
    if (complete) {
      b = handler.writeBegin();
    }

    if (b) {

      lti::write(handler,"variance",variance);
      lti::write(handler,"kernelSize",kernelSize);
      lti::write(handler,"maximumCorners",maximumCorners);
      lti::write(handler,"scale",scale);
      lti::write(handler,"localMaximaParameters",localMaximaParameters);

      switch(gradientMode) {
        case HarrisGradient:
          lti::write(handler,"gradientMode","HarrisGradient");
          break;
        case OptimalGradient3:
          lti::write(handler,"gradientMode","OptimalGradient3");
          break;
        case OptimalGradient5:
          lti::write(handler,"gradientMode","OptimalGradient5");
          break;
        default:
          lti::write(handler,"gradientMode","HarrisGradient");
      }
    }

# ifndef _LTI_MSC_VER
    // This is the standard C++ code, which MS Visual C++ 6 is not able to
    // compile...
    b = b && cornerDetector::parameters::write(handler,false);
# else
    bool (cornerDetector::parameters::* p_writeMS)(ioHandler&,const bool&) const =
      cornerDetector::parameters::writeMS;
    b = b && (this->*p_writeMS)(handler,false);
# endif

    if (complete) {
      b = b && handler.writeEnd();
    }

    return b;
  }

# ifdef _LTI_MSC_VER
  bool harrisCorners::parameters::write(ioHandler& handler,
                                         const bool& complete) const {
    // ...we need this workaround to cope with another really awful MSVC bug.
    return writeMS(handler,complete);
  }
# endif

  /*
   * read the parameters from the given ioHandler
   * @param handler the ioHandler to be used
   * @param complete if true (the default) the enclosing begin/end will
   *        be also read, otherwise only the data block will be read.
   * @return true if write was succeful
   */
# ifndef _LTI_MSC_VER
  bool harrisCorners::parameters::read(ioHandler& handler,
                                        const bool& complete)
# else
  bool harrisCorners::parameters::readMS(ioHandler& handler,
                                          const bool& complete)
# endif
  {
    bool b = true;
    if (complete) {
      b = handler.readBegin();
    }

    if (b) {

      lti::read(handler,"variance",variance);
      lti::read(handler,"kernelSize",kernelSize);
      lti::read(handler,"maximumCorners",maximumCorners);
      lti::read(handler,"scale",scale);
      lti::read(handler,"localMaximaParameters",localMaximaParameters);

      std::string str;
      lti::read(handler,"gradientMode",str);
      if (str == "HarrisGradient") {
        gradientMode = HarrisGradient;
      } else if (str == "OptimalGradient3") {
        gradientMode = OptimalGradient3;
      } else if (str == "OptimalGradient5") {
        gradientMode = OptimalGradient5;
      } else {
        gradientMode = HarrisGradient;
      }

    }

# ifndef _LTI_MSC_VER
    // This is the standard C++ code, which MS Visual C++ 6 is not able to
    // compile...
    b = b && cornerDetector::parameters::read(handler,false);
# else
    bool (cornerDetector::parameters::* p_readMS)(ioHandler&,const bool&) =
      cornerDetector::parameters::readMS;
    b = b && (this->*p_readMS)(handler,false);
# endif

    if (complete) {
      b = b && handler.readEnd();
    }

    return b;
  }

# ifdef _LTI_MSC_VER
  bool harrisCorners::parameters::read(ioHandler& handler,
                                        const bool& complete) {
    // ...we need this workaround to cope with another really awful MSVC
    // bug.
    return readMS(handler,complete);
  }
# endif

  // --------------------------------------------------
  // harrisCorners
  // --------------------------------------------------

  // default constructor
  harrisCorners::harrisCorners()
    : cornerDetector(){

    // create an instance of the parameters with the default values
    parameters defaultParameters;
    // set the default parameters
    setParameters(defaultParameters);
  }

  // default constructor
  harrisCorners::harrisCorners(const parameters& par)
    : cornerDetector() {
    // set the given parameters
    setParameters(par);
  }


  // copy constructor
  harrisCorners::harrisCorners(const harrisCorners& other)
    : cornerDetector()  {
    copy(other);
  }

  // destructor
  harrisCorners::~harrisCorners() {
  }

  // returns the name of this type
  const char* harrisCorners::getTypeName() const {
    return "harrisCorners";
  }

  // which apply is faster
  functor::eFasterApply harrisCorners::theFaster() const {
    return OnCopy;
  }

  // copy member
  harrisCorners&
  harrisCorners::copy(const harrisCorners& other) {
    cornerDetector::copy(other);
    return (*this);
  }

  // alias for copy member
  harrisCorners&
    harrisCorners::operator=(const harrisCorners& other) {
    return (copy(other));
  }


  // clone member
  functor* harrisCorners::clone() const {
    return new harrisCorners(*this);
  }

  // return parameters
  const harrisCorners::parameters&
    harrisCorners::getParameters() const {
    const parameters* par =
      dynamic_cast<const parameters*>(&functor::getParameters());
    if(isNull(par)) {
      throw invalidParametersException(getTypeName());
    }
    return *par;
  }

  // -------------------------------------------------------------------
  // The apply-methods!
  // -------------------------------------------------------------------

  // On place apply for type channel8!
  bool harrisCorners::apply(channel8& srcdest) const {
    return cornerDetector::apply(srcdest);
  };

  // On place apply for type channel!
  bool harrisCorners::apply(channel& srcdest) const {
    return cornerDetector::apply(srcdest);
  };

  // On copy apply for type channel8!
  bool harrisCorners::apply(const channel8& src,channel8& dest) const {
    pointList pts;
    if (apply(src,pts)) {
      const parameters& par = getParameters();
      pointList::iterator it;
      dest.resize(src.size(),par.noCornerValue,false,true);
      for (it=pts.begin();it!=pts.end();++it) {
        dest.at(*it)=par.cornerValue;
      }
    }

    return false;
  };

  // On copy apply for type channel!
  bool harrisCorners::apply(const channel& src,channel& dest) const {
    const parameters& param = getParameters();

    channel gx,gy,fxx,fxy,fyy,tmp;
    float maxCornerness;

    pointList cornerMax;

    getGradient(src,gx,gy);
    getSecondOrder(gx,gy,fxx,fxy,fyy);
    getCornerness(fxx,fxy,fyy,param.scale,tmp,maxCornerness);
    findCornerMaxima(tmp,dest,cornerMax);

    return true;
  };

  /*
   * operates on a copy of the given %parameters.
   * @param src channel8 with the source data.
   * @param dest list of corners
   * @return true if apply successful or false otherwise.
   */
  bool harrisCorners::apply(const channel8& src,pointList& dest) const {
    channel tmp;
    tmp.castFrom(src);
    return apply(tmp,dest);
  }

  /*
   * operates on a copy of the given %parameters.
   * @param src channel with the source data.
   * @param dest list of corners
   * @return true if apply successful or false otherwise.
   */
  bool harrisCorners::apply(const channel& src,pointList& dest) const {
    const parameters& param = getParameters();

    channel gx,gy,fxx,fxy,fyy,tmp,d;
    float maxCornerness;

    getGradient(src,gx,gy);
    getSecondOrder(gx,gy,fxx,fxy,fyy);
    getCornerness(fxx,fxy,fyy,param.scale,tmp,maxCornerness);
    return findCornerMaxima(tmp,d,dest);
  };

  /*
   * operates on a copy of the given %parameters.
   * @param src channel with the source data.
   * @param dest list of corners
   * @return true if apply successful or false otherwise.
   */
  bool harrisCorners::apply(const channel& src,
                            channel& cornerness,
                            float& maxCornerness) const {
    const parameters& param = getParameters();

    channel gx,gy,fxx,fxy,fyy,d;

    getGradient(src,gx,gy);
    getSecondOrder(gx,gy,fxx,fxy,fyy);
    return getCornerness(fxx,fxy,fyy,param.scale,cornerness,maxCornerness);
  };

  bool harrisCorners::apply(const channel& src,
                            channel& cornerness,
                            float& maxCornerness,
                            pointList& dest) const {
    const parameters& param = getParameters();

    channel gx,gy,fxx,fxy,fyy,d;

    getGradient(src,gx,gy);
    getSecondOrder(gx,gy,fxx,fxy,fyy);
    getCornerness(fxx,fxy,fyy,param.scale,cornerness,maxCornerness);
    return findCornerMaxima(cornerness,d,dest);
  };

  bool harrisCorners::setUpConvolution(convolution& filterX,
                                       convolution& filterY) const {

    const parameters& param = getParameters();
    convolution::parameters filterPar;
    filterPar.boundaryType = lti::Constant;

    // create the required convolution functors to compute the gradient
    if (param.gradientMode == parameters::HarrisGradient) {
      // construct the harris kernel [-2,-1,0,+1,+2]
      kernel1D<float> h1(-2,2,0.0f);
      for (int i=-2;i<3;++i) {
        h1.at(i)=static_cast<float>(i);
      }
      kernel1D<float> h2(0,0,1.0f);

      sepKernel<float> sh;
      sh.setNumberOfPairs(1);

      // the X kernel
      sh.getColFilter(0).copy(h1);
      sh.getRowFilter(0).copy(h2);
      filterPar.setKernel(sh);
      filterX.setParameters(filterPar);

      // the Y kernel
      sh.getColFilter(0).copy(h2);
      sh.getRowFilter(0).copy(h1);
      filterPar.setKernel(sh);
      filterY.setParameters(filterPar);
    } else if (param.gradientMode == parameters::OptimalGradient3) {
      gradientKernelX<float> gx(3);
      gradientKernelY<float> gy(3);
      filterPar.setKernel(gx);
      filterX.setParameters(filterPar);
      filterPar.setKernel(gy);
      filterY.setParameters(filterPar);
    } else if (param.gradientMode == parameters::OptimalGradient5) {
      gradientKernelX<float> gx(5);
      gradientKernelY<float> gy(5);
      filterPar.setKernel(gx);
      filterX.setParameters(filterPar);
      filterPar.setKernel(gy);
      filterY.setParameters(filterPar);
    } else {
      setStatusString("Unknown gradient kernel type");
      return false;
    }

    return true;
  }

  /**
   * compute the simplest gradient of the given src image
   */
  bool harrisCorners::getGradient(const channel& src,
                                        channel& gradX,
                                        channel& gradY) const {

    convolution filterX,filterY;
    return (setUpConvolution(filterX,filterY) &&
            filterX.apply(src,gradX) &&
            filterY.apply(src,gradY));

  }

  /*
   * compute the second order
   */
  bool harrisCorners::getSecondOrder(const channel& gradx,
                                     const channel& grady,
                                     channel& fxx,
                                     channel& fxy,
                                     channel& fyy) const {

    const parameters& par = getParameters();

    channel gradxx,gradyy;
    gaussKernel2D<float> gk(par.kernelSize,par.variance);
    convolution filter;
    convolution::parameters filterPar;
    filterPar.boundaryType = lti::Constant;
    filterPar.setKernel(gk);
    filter.setParameters(filterPar);

    fxx.emultiply(gradx,gradx);
    fxy.emultiply(gradx,grady);
    fyy.emultiply(grady,grady);

    return (filter.apply(fxx) && filter.apply(fxy) && filter.apply(fyy));
  }


  /*
   * compute cornerness
   */
  bool harrisCorners::getCornerness(const channel& fxx,
                                    const channel& fxy,
                                    const channel& fyy,
                                    const float scale,
                                    channel& cornerness,
                                    float& maxCornerness) const {
    // we can assume that all channels are connected, but try it out if not
    if ((fxx.getMode() != channel::Connected) ||
        (fxy.getMode() != channel::Connected) ||
        (fyy.getMode() != channel::Connected)) {
      setStatusString("Channels not contigous in getCornerness");
      return false;
    }

    int i;
    const int end = fxx.rows()*fxx.columns();
    const float *const pfxx = &fxx.at(0);
    const float *const pfxy = &fxy.at(0);
    const float *const pfyy = &fyy.at(0);

    cornerness.resize(fxx.size(),0,false,false);
    float* pcor = &cornerness.at(0);

    float det,trace,txx,txy,tyy,c;
    float maxc = 0.0f;

    for (i=0;i<end;++i) {
      txx=pfxx[i];
      txy=pfxy[i];
      tyy=pfyy[i];
      det=txx*tyy - txy*txy;
      trace=txx+tyy;
      c = det-scale*trace*trace;
      pcor[i]=c;
      if (c>maxc) {
        maxc=c;
      }
    }

    maxCornerness = maxc;
    return true;
  }

  /*
   * find corners with maximal cornerness
   */
  bool harrisCorners::findCornerMaxima(const channel& cornerness,
                                             channel& cornersOnly,
                                             pointList& cornerMax) const {
    const parameters& par = getParameters();

    const float corner = par.cornerValue/255.0f;
    const float noCorner = par.noCornerValue/255.0f;

    localMaxima<float> lmax;
    localMaxima<float>::parameters lmaxPar(par.localMaximaParameters);
    lmaxPar.noMaxValue = noCorner;
    lmax.setParameters(lmaxPar);

    if (lmax.apply(cornerness,cornersOnly,cornerMax)) {
      pointList::iterator it;
      int i;
      for (it=cornerMax.begin(),i=0;
           (it!=cornerMax.end()) && (i<par.maximumCorners);
           ++it,++i) {
        cornersOnly.at(*it) = corner;
      }

      for (;it!=cornerMax.end();++it) {
        cornersOnly.at(*it) = noCorner;
      }

      return true;
    }
    return false;
  }

}




