/*
 * Copyright (C) 1998, 1999, 2000, 2001
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 *
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.
 */


/*--------------------------------------------------------------------
 * project ....: LTI-Lib: Image Processing and Computer Vision Library
 * file .......: ltiHarrisCorners.h
 * authors ....: Pablo Alvarado
 * organization: LTI, RWTH Aachen
 * creation ...: 31.10.2002
 * revisions ..: $Id: ltiHarrisCorners.h,v 1.8 2003/11/19 06:55:13 alvarado Exp $
 */

#ifndef _LTI_HARRIS_CORNERS_H_
#define _LTI_HARRIS_CORNERS_H_

#include "ltiCornerDetector.h"
#include "ltiPointList.h"
#include "ltiConvolution.h"
#include "ltiLocalMaxima.h"

namespace lti {
  /**
   * Harris Corner Detector.
   *
   * The Harris corner detector is based on
   *
   * C. Harris and M. Stephens, "A Combined Corner and Edge Detector",
   * in Proc. 4th Alvey Vision Conference, pp. 147-151, 1988
   *
   * The corners are defined as the local maxima of a cornerness
   * channel.  The cornerness \a c(x,y) of a channel \a f(x,y) is
   * defined as:
   *
   * \f[ c(x,y)=(f_{xx}f_{yy} - f_{xy}^2) - s(f_{xx}+f_{yy})^2 \f]
   *
   * with \f$f_{\zeta\xi}\f$ the second derivative of the channel
   * respect \f$\zeta\f$ and \f$\xi\f$.
   *
   * Note that the first term of the cornerness represents the determinant
   * of the Hessian matrix, and the second term its trace multiplied by
   * a scaling factor \e s.  Remember that the product of the eigenvalues 
   * equals the determinant of the matrix and the sum of eigenvalues equals
   * its trace.
   *
   * Slighly modifications are used in order to increase the
   * quality of the detection.
   */
  class harrisCorners : public cornerDetector {
  public:
    /**
     * the parameters for the class harrisCorners
     */
    class parameters : public cornerDetector::parameters {
    public:
      /**
       * default constructor
       */
      parameters();

      /**
       * copy constructor
       * @param other the parameters object to be copied
       */
      parameters(const parameters& other);

      /**
       * destructor
       */
      ~parameters();

      /**
       * returns name of this type
       */
      const char* getTypeName() const;

      /**
       * copy the contents of a parameters object
       * @param other the parameters object to be copied
       * @return a reference to this parameters object
       */
      parameters& copy(const parameters& other);

      /**
       * copy the contents of a parameters object
       * @param other the parameters object to be copied
       * @return a reference to this parameters object
       */
      parameters& operator=(const parameters& other);


      /**
       * returns a pointer to a clone of the parameters
       */
      virtual functor::parameters* clone() const;

      /**
       * write the parameters in the given ioHandler
       * @param handler the ioHandler to be used
       * @param complete if true (the default) the enclosing begin/end will
       *        be also written, otherwise only the data block will be written.
       * @return true if write was successful
       */
      virtual bool write(ioHandler& handler,const bool& complete=true) const;

      /**
       * read the parameters from the given ioHandler
       * @param handler the ioHandler to be used
       * @param complete if true (the default) the enclosing begin/end will
       *        be also written, otherwise only the data block will be written.
       * @return true if write was successful
       */
      virtual bool read(ioHandler& handler,const bool& complete=true);

#     ifdef _LTI_MSC_VER
      /**
       * this function is required by MSVC only, as a workaround for a
       * very awful bug, which exists since MSVC V.4.0, and still by
       * V.6.0 with all bugfixes (so called "service packs") remains
       * there...  This method is also public due to another bug, so please
       * NEVER EVER call this method directly: use read() instead
       */
      bool readMS(ioHandler& handler,const bool& complete=true);

      /**
       * this function is required by MSVC only, as a workaround for a
       * very awful bug, which exists since MSVC V.4.0, and still by
       * V.6.0 with all bugfixes (so called "service packs") remains
       * there...  This method is also public due to another bug, so please
       * NEVER EVER call this method directly: use write() instead
       */
      bool writeMS(ioHandler& handler,const bool& complete=true) const;
#     endif

      // ------------------------------------------------
      // the parameters
      // ------------------------------------------------

      /**
       * Variance used for the Gaussian kernel used as window for the
       * autocorrelation.
       *
       * Default value : -1 (see lti::gaussKernel1D<float>::generate() for
       *                 the meaning of this)
       */
      float variance;

      /**
       * size of the Gaussian kernel used as window for the autocorrelation
       *
       * Default value: 7
       */
      int kernelSize;

      /**
       * Maximum number of corners to be detected.
       *
       * Default value: 300
       */
      int maximumCorners;

      /**
       * Scale factor in Harris auto-correlation
       *
       * Default value: 0.04f
       */
      float scale;

      /**
       * A localMaxima<float> %functor is used to detect the corners.
       * The parameters for it can be given here.
       *
       * Default value: localMaxima<float>::parameters()
       */
      localMaxima<float>::parameters localMaximaParameters;

      /**
       * types of gradient type used
       */
      enum eGradientMode {
        HarrisGradient,   /**< Original Harris gradient kernels are
                           *   [-2 -1 0 1 2]
                           */
        OptimalGradient3, /**< Gradient Kernel with size 3x3
                           *   See also: lti::gradientKernelX
                           */
        OptimalGradient5  /**< Gradient Kernel with size 5x5
                           *   See also: lti::gradientKernelX
                           */
      };

      /**
       * Mode used to compute the gradient.
       *
       * Default: HarrisGradient
       */
      eGradientMode gradientMode;
    };

    /**
     * default constructor
     */
    harrisCorners();

    /**
     * Construct a functor using the given parameters
     */
    harrisCorners(const parameters& par);

    /**
     * copy constructor
     * @param other the object to be copied
     */
    harrisCorners(const harrisCorners& other);

    /**
     * destructor
     */
    virtual ~harrisCorners();

    /**
     * returns the name of this type ("harrisCorners")
     */
    virtual const char* getTypeName() const;


    /**
     * Extract the corners in srcdest and leave the result there too.
     *
     * The result is a mask containing only the corners.  The background
     * will be marked using parameters::noCornerValue and the corners will
     * have the value parameters::cornerValue (these are inherited from
     * the parent class).
     *
     * @param srcdest channel8 with the source data.  The result
     *                 will be left here too.
     * @return true if apply successful or false otherwise.
     */
    virtual bool apply(channel8& srcdest) const;

    /**
     * Extract the corners in srcdest and leave the result there too.
     *
     * The result is a mask containing only the corners.  The background
     * will be marked using parameters::noCornerValue/255 and the corners will
     * have the value parameters::cornerValue/255 (these are inherited from
     * the parent class).
     *
     * @param srcdest channel with the source data.  The result
     *                 will be left here too.
     * @return true if apply successful or false otherwise.
     */
    virtual bool apply(channel& srcdest) const;

    /**
     * Extract the corners in src and leave the result in dest.
     *
     * The result is a mask containing only the corners.  The background
     * will be marked using parameters::noCornerValue and the corners will
     * have the value parameters::cornerValue (these are inherited from
     * the parent class).
     *
     * @param src channel8 with the source data.
     * @param dest channel8 where the result will be left.
     * @return true if apply successful or false otherwise.
     */
    virtual bool apply(const channel8& src,channel8& dest) const;

    /**
     * Extract the corners in src and leave the result in dest.
     *
     * The result is a mask containing only the corners.  The background
     * will be marked using parameters::noCornerValue/255 and the corners will
     * have the value parameters::cornerValue/255 (these are inherited from
     * the parent class).
     *
     * @param src channel with the source data.
     * @param dest channel where the result will be left.
     * @return true if apply successful or false otherwise.
     */
    virtual bool apply(const channel& src,channel& dest) const;

    /**
     * Extract the corners in src and leave them in dest.
     *
     * @param src channel8 with the source data.
     * @param dest list of corners
     * @return true if apply successful or false otherwise.
     */
    virtual bool apply(const channel8& src,pointList& dest) const;

    /**
     * Extract the corners in src and leave them in dest.
     *
     * @param src channel with the source data.
     * @param dest list of corners
     * @return true if apply successful or false otherwise.
     */
    virtual bool apply(const channel& src,pointList& dest) const;

    /**
     * Extract the cornerness channel belonging to src.
     *
     * Note that the cornerness is in principle a partial result.
     *
     * @param src channel with the source data.
     * @param cornerness channel containing the cornerness of the source data.
     * @param maxCornerness maximum value found in cornerness.
     * @param dest list containing the position of all detected corners.
     * @return true if apply successful or false otherwise.
     */
    virtual bool apply(const channel& src,
                       channel& cornerness,
                       float& maxCornerness,
                       pointList& dest) const;

    /**
     * Extract the cornerness channel belonging to src.
     *
     * Note that this is in principle a partial result.  It would require
     * further processing to get the corners, but some other functors could
     * make use of this.
     *
     * @param src channel with the source data.
     * @param cornerness channel containing the cornerness of the source data.
     * @param maxCornerness maximum value found in cornerness.
     * @return true if apply successful or false otherwise.
     */
    virtual bool apply(const channel& src,
                       channel& cornerness,
                       float& maxCornerness) const;

    /**
     * returns whether the onPlace version of apply is faster or
     * the onCopy version.
     */
    virtual functor::eFasterApply theFaster() const;

    /**
     * copy data of "other" functor.
     * @param other the functor to be copied
     * @return a reference to this functor object
     */
    harrisCorners& copy(const harrisCorners& other);

    /**
     * alias for copy member
     * @param other the functor to be copied
     * @return a reference to this functor object
     */
    harrisCorners& operator=(const harrisCorners& other);

    /**
     * returns a pointer to a clone of this functor.
     */
    virtual functor* clone() const;

    /**
     * returns used parameters
     */
    const parameters& getParameters() const;

  protected:

    /**
     * setup the convolution filters with the proper kernels
     * for gradient computation
     */
    bool setUpConvolution(convolution& filterX,
                          convolution& filterY) const;

    /**
     * compute the simplest gradient of the given src image
     */
    bool getGradient(const channel& src,
                           channel& gradX,
                           channel& gradY) const;

    /**
     * compute the second order
     */
    bool getSecondOrder(const channel& gradx,
                        const channel& grady,
                        channel& fxx,
                        channel& fxy,
                        channel& fyy) const;

    /**
     * compute cornerness
     */
    bool getCornerness(const channel& fxx,
                       const channel& fxy,
                       const channel& fyy,
                       const float scale,
                       channel& cornerness,
                       float& maxCornerness) const;

    /**
     * find corners with maximal cornerness
     */
    bool findCornerMaxima(const channel& cornerness,
                          channel& cornersOnly,
                          pointList& cornerMax) const;

  };
}

#endif

