/*
 * Copyright (C) 1998, 1999, 2000, 2001
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 *
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.
 */


/*----------------------------------------------------------------
 * project ....: LTI Digital Image/Signal Processing Library
 * file .......: ltiMergeHSVToImage.cpp
 * authors ....: Pablo Alvarado, Stefan Syberichs, Thomas Rusert
 * organization: LTI, RWTH Aachen
 * creation ...: 19.04.99
 * revisions ..: $Id: ltiMergeHSVToImage.cpp,v 1.1 2003/11/14 16:07:21 alvarado Exp $
 */

#include "ltiMergeHSVToImage.h"

namespace lti {

  // merge HSV channels to Image

  // return the name of this type
  const char* mergeHSVToImage::getTypeName() const {
    return "mergeHSVToImage";
  }

  // returns a pointer to a clone of the functor.
  functor* mergeHSVToImage::clone() const {
    return (new mergeHSVToImage(*this));
  }

  // merge float channels
  bool mergeHSVToImage::apply(const matrix<float>& c1,
                              const matrix<float>& c2,
                              const matrix<float>& c3,
                              image& img) const {
    int xSize,ySize;
    point p;
    rgbPixel pix;
    int i,r,q,t,v; // auxiliary for transformation
    float f;       // auxiliary for transformation
    float H, S, V; // channels

    xSize=c1.columns();
    ySize=c1.rows();
    if ((c2.columns()!=xSize)||
	(c2.rows()!=ySize)||
	(c3.columns()!=xSize)||
	(c3.rows()!=ySize)) {
      setStatusString("sizes of channels do not match");
      return false;
    }

    img.resize(ySize,xSize,rgbPixel(),false,false);

    // H = c1
    // S = c2
    // I = c3
    for (p.y=0;p.y<ySize;p.y++)
      for (p.x=0;p.x<xSize;p.x++) {

	H = c1.at(p);
	S = c2.at(p);
	V = c3.at(p);

	if (c2.at(p)==0) {
	  pix.setRed((ubyte)(V*255.0f));
	  pix.setGreen((ubyte)(V*255.0f));
	  pix.setBlue((ubyte)(V*255.0f));

	}
	else {
	  i = (int)(H*6);
	  f = H*6-i;
	  r =(int) ((V*(1.0f-S))*255.0f);
	  q =(int) ((V*(1.0f-(S*f)))*255.0f);
	  t =(int)((V*(1.0f-(S*(1.0f-f))))*255.0f);
	  v =(int)(V*255.0f);

	  switch (i) {
	  case 0:
	  case 6: 	pix.setRed(v);
	    pix.setGreen(t);
	    pix.setBlue(r);
	    break;

	  case 1:		pix.setRed(q);
	    pix.setGreen(v);
	    pix.setBlue(r);
	    break;

	  case 2:		pix.setRed(r);
	    pix.setGreen(v);
	    pix.setBlue(t);
	    break;

	  case 3:		pix.setRed(r);
	    pix.setGreen(q);
	    pix.setBlue(v);
	    break;

	  case 4:		pix.setRed(t);
	    pix.setGreen(r);
	    pix.setBlue(v);
	    break;

	  case 5:		pix.setRed(v);
	    pix.setGreen(r);
	    pix.setBlue(q);
	    break;
	  }
	}

	img.at(p) = pix;
      }

    return true;
  };


  // merge 8-bit channels channels
  bool mergeHSVToImage::apply(const channel8& c1,
                              const channel8& c2,
                              const channel8& c3,
                              image& img) const {
    int xSize,ySize;
    point p;
    rgbPixel pix;
    int i,r,q,t,v;     // auxiliary for transformation
    float f;           // auxuliary for transformation
    float H, S, V;     // channels

    xSize=c1.columns();
    ySize=c1.rows();
    if ((c2.columns()!=xSize)||
	(c2.rows()!=ySize)||
	(c3.columns()!=xSize)||
	(c3.rows()!=ySize)) {
      setStatusString("sizes of channels do not match");
      return false;
    }

    img.resize(ySize,xSize,rgbPixel(),false,false);

    // H = c1
    // S = c2
    // I = c3

    for (p.y=0;p.y<ySize;p.y++)
      for (p.x=0;p.x<xSize;p.x++) {

	H = ((float)(c1.at(p)))/255.0f;
	S = ((float)(c2.at(p)))/255.0f;
	V = ((float)(c3.at(p)))/255.0f;

	if (S==0) {
	  pix.setRed((ubyte)(V*255.0f));
	  pix.setGreen((ubyte)(V*255.0f));
	  pix.setBlue((ubyte)(V*255.0f));

	}
	else {
	  i = (int)(H*6);
	  f = H*6-i;
	  r =(int) ((V*(1.0f-S))*255.0f);
	  q =(int) ((V*(1.0f-(S*f)))*255.0f);
	  t = (int)((V*(1.0f-(S*(1.0f-f))))*255.0f);
	  v = (int)(V*255.0f);

	  switch (i) {
	  case 0:
	  case 6: 	pix.setRed(v);
	    pix.setGreen(t);
	    pix.setBlue(r);
	    break;

	  case 1:		pix.setRed(q);
	    pix.setGreen(v);
	    pix.setBlue(r);
	    break;

	  case 2:		pix.setRed(r);
	    pix.setGreen(v);
	    pix.setBlue(t);
	    break;

	  case 3:		pix.setRed(r);
	    pix.setGreen(q);
	    pix.setBlue(v);
	    break;

	  case 4:		pix.setRed(t);
	    pix.setGreen(r);
	    pix.setBlue(v);
	    break;

	  case 5:		pix.setRed(v);
	    pix.setGreen(r);
	    pix.setBlue(q);
	    break;
	  }
	}

	img.at(p) = pix;
      }

    return true;
  };

  // merge float values
  bool mergeHSVToImage::apply(const float& H,
                              const float& S,
                              const float& V,
                              rgbPixel& pixel) const {

    int i,r,q,t,v; // auxiliary for transformation
    float f;       // auxiliary for transformation

    if (S==0) {
      pixel.setRed((ubyte)(V*255.0f));
      pixel.setGreen((ubyte)(V*255.0f));
      pixel.setBlue((ubyte)(V*255.0f));
    }
    else {
      i = (int)(H*6);
      f = H*6-i;
      r =(int) ((V*(1.0f-S))*255.0f);
      q =(int) ((V*(1.0f-(S*f)))*255.0f);
      t =(int)((V*(1.0f-(S*(1.0f-f))))*255.0f);
      v =(int)(V*255.0f);

      switch (i) {
	case 0:
	case 6:
	  pixel.setRed(v);
	  pixel.setGreen(t);
	  pixel.setBlue(r);
	  break;
	case 1:
	  pixel.setRed(q);
	  pixel.setGreen(v);
	  pixel.setBlue(r);
	  break;
	case 2:
	  pixel.setRed(r);
	  pixel.setGreen(v);
	  pixel.setBlue(t);
	  break;
	case 3:
	  pixel.setRed(r);
	  pixel.setGreen(q);
	  pixel.setBlue(v);
	  break;
	case 4:
	  pixel.setRed(t);
	  pixel.setGreen(r);
	  pixel.setBlue(v);
	  break;
	case 5:
	  pixel.setRed(v);
	  pixel.setGreen(r);
	  pixel.setBlue(q);
	  break;
      }
    }

  return true;
  }

  // merge 8-bit values
  bool mergeHSVToImage::apply(const ubyte& c1,
                              const ubyte& c2,
                              const ubyte& c3,
                              rgbPixel& pixel) const {

    int i,r,q,t,v;     // auxiliary for transformation
    float f;           // auxuliary for transformation
    float H, S, V;     // channels

    H = float(c1)/255.0f;
    S = float(c2)/255.0f;
    V = float(c3)/255.0f;

    if (S==0) {
      pixel.setRed((ubyte)(V*255.0f));
      pixel.setGreen((ubyte)(V*255.0f));
      pixel.setBlue((ubyte)(V*255.0f));
    }
    else {
      i = (int)(H*6);
      f = H*6-i;
      r =(int) ((V*(1.0f-S))*255.0f);
      q =(int) ((V*(1.0f-(S*f)))*255.0f);
      t = (int)((V*(1.0f-(S*(1.0f-f))))*255.0f);
      v = (int)(V*255.0f);

      switch (i) {
	case 0:
	case 6:
	  pixel.setRed(v);
	  pixel.setGreen(t);
	  pixel.setBlue(r);
	  break;
	case 1:
	  pixel.setRed(q);
	  pixel.setGreen(v);
	  pixel.setBlue(r);
	  break;
	case 2:
	  pixel.setRed(r);
	  pixel.setGreen(v);
	  pixel.setBlue(t);
	  break;
	case 3:
	  pixel.setRed(r);
	  pixel.setGreen(q);
	  pixel.setBlue(v);
	  break;
	case 4:
	  pixel.setRed(t);
	  pixel.setGreen(r);
	  pixel.setBlue(v);
	  break;
	case 5:
	  pixel.setRed(v);
	  pixel.setGreen(r);
	  pixel.setBlue(q);
	  break;
      }
    }

    return true;
  };

} // end of namespace

