/*
 * Copyright (C) 1998, 1999, 2000, 2001
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 *
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.
 */


/*----------------------------------------------------------------
 * project ....: LTI Digital Image/Signal Processing Library
 * file .......: ltiOgdKernels_template.h
 * authors ....: Pablo Alvarado
 * organization: LTI, RWTH Aachen
 * creation ...: 28.04.00
 * revisions ..: $Id: ltiOgdKernels_template.h,v 1.7 2003/05/16 13:21:38 alvarado Exp $
 */

#include "ltiMath.h"
#include "ltiGaussKernels.h"

namespace lti {
  // first order ogd

  // default constructor
  template<class T>
    ogd1Kernel<T>::ogd1Kernel() : sepKernel<T>() {
  }

  // constructor
  template<class T>
    ogd1Kernel<T>::ogd1Kernel(const int& size,
                              const double& theVariance,
                              const double& theAngle)
    : sepKernel<T>() {
    generate(size,theVariance,theAngle);
  }

  // generate filter
  template<class T>
    void  ogd1Kernel<T>::generate(const int& size,
                                  const double& theVariance,
                                  const double& theAngle) {

    gaussKernel1D<T> gauss(size,theVariance);

    setNumberOfPairs(2);

    getRowFilter(0).copy(gauss);
    getColFilter(0).copy(gauss);
    getRowFilter(1).copy(gauss);
    getColFilter(1).copy(gauss);

    int x;
    double cost,sint;

    cost=cos(theAngle);
    sint=sin(theAngle);

    for (x  = gauss.firstIdx();
         x <= gauss.lastIdx();
         x++) {
      getRowFilter(0).at(x) *= static_cast<T>(-x/theVariance);
      getColFilter(0).at(x) *= static_cast<T>(cost);
      getRowFilter(1).at(x) *= static_cast<T>(sint);
      getColFilter(1).at(x) *= static_cast<T>(-x/theVariance);
    }
  }

  template<class T>
    void  ogd1Kernel<T>::generateBasisKernel(const int& whichBasis,
                                             const int& size,
                                             const double& theVariance) {

    gaussKernel1D<T> gauss(size,theVariance);

    setNumberOfPairs(1);

    getRowFilter(0).copy(gauss);
    getColFilter(0).copy(gauss);

    int x;
    kernel1D<T>* f = 0;

    if (whichBasis == 0) {
      f = &getRowFilter(0);
    } else {
      f = &getColFilter(0);
    }

    for (x  = gauss.firstIdx();
         x <= gauss.lastIdx();
         x++) {
      f->at(x) *= static_cast<T>(-x/theVariance);
    }
  }

  // second order filter kernel

  // constructor
  template<class T>
    ogd2Kernel<T>::ogd2Kernel()
    : sepKernel<T>() {
  }

  template<class T>
    ogd2Kernel<T>::ogd2Kernel(const int& size,
                              const double& theVariance,
                              const double& theAngle)
    : sepKernel<T>() {
    generate(size,theVariance,theAngle);
  }

  // generate filter
  template<class T>
    void  ogd2Kernel<T>::generate(const int& size,
                                  const double& theVariance,
                                  const double& theAngle) {

    gaussKernel1D<T> gauss(size,theVariance);

    setNumberOfPairs(3);

    getRowFilter(0).copy(gauss);
    getColFilter(0).copy(gauss);
    getRowFilter(1).copy(gauss);
    getColFilter(1).copy(gauss);
    getRowFilter(2).copy(gauss);
    getColFilter(2).copy(gauss);

    int x;
    double cost,sint;

    cost=cos(theAngle);
    sint=sin(theAngle);

    for (x  = gauss.firstIdx();
         x <= gauss.lastIdx();
         x++) {
      getRowFilter(0).at(x) *= static_cast<T>(((x*x)-theVariance)/
                                              (theVariance*theVariance));
      getColFilter(1).at(x) *= static_cast<T>(((x*x)-theVariance)/
                                              (theVariance*theVariance));
      getRowFilter(2).at(x) *= static_cast<T>(x/theVariance);
      getColFilter(2).at(x) *= static_cast<T>(x/theVariance);
    }
  }

  template<class T>
    void  ogd2Kernel<T>::generateBasisKernel(const int& basis,
                                             const int& size,
                                             const double& theVariance) {
    gaussKernel1D<T> gauss(size,theVariance);

    setNumberOfPairs(1);

    getRowFilter(0).copy(gauss);
    getColFilter(0).copy(gauss);

    int x;
    kernel1D<T>* f = 0;

    if (basis == 0) {
      f = &getRowFilter(0);
    } else if (basis == 1) {
      f = &getColFilter(0);
    }

    if (basis < 2) {
      for (x  = gauss.firstIdx();
           x <= gauss.lastIdx();
           x++) {
        f->at(x) *= static_cast<T>(((x*x)-theVariance)/
                                   (theVariance*theVariance));
      }
    } else {
      for (x = gauss.firstIdx();
           x <= gauss.lastIdx();
           x++) {
        getRowFilter(0).at(x) *= static_cast<T>(x/theVariance);
        getColFilter(0).at(x) *= static_cast<T>(x/theVariance);
      }
    }
  }
}

