/*
 * Copyright (C) 1998, 1999, 2000, 2001
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 *
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.
 */


/*----------------------------------------------------------------
 * project ....: LTI Digital Image/Signal Processing Library
 * file .......: ltiEquationSystem_template.h
 * authors ....: Thomas Rusert
 * organization: LTI, RWTH Aachen
 * creation ...: 02.06.99
 * revisions ..: $Id: ltiEquationSystem_template.h,v 1.13 2003/05/16 07:26:33 alvarado Exp $
 */

#include "ltiMath.h"
#include "ltiMatrixDecomposition.h"
#include <limits>

namespace lti {
  //--- decompositionSolution ---
  //-----------------------------

  // construction
  template<class T>
  decompositionSolution<T>::decompositionSolution() {
    decomposed=false;
  }

  template<class T>
  decompositionSolution<T>::decompositionSolution(const parameters& theParams){
    setParameters(theParams);
  }

  template<class T>
  decompositionSolution<T>::decompositionSolution(const matrix<T>& theMatrix) {
    parameters tmpParam;
    tmpParam.systemMatrix = theMatrix;
    setParameters(tmpParam);
  }

  // sets the functor's parameters.
  template<class T>
  bool decompositionSolution<T>::setParameters(
                             const functor::parameters& theParam) {
    decomposed=false;
    return linearEquationSystemSolutionMethod<T>::setParameters(theParam);
  }

  // returns the current parameters
  template<class T>
  const typename decompositionSolution<T>::parameters&
  decompositionSolution<T>::getParameters() const {
    const parameters* params =
      dynamic_cast<const parameters*>(&functor::getParameters());
    if(isNull(params)) {
      throw functor::invalidParametersException(getTypeName());
    }
    return *params;
  }

  // copy data of "other" functor.
  template<class T>
  decompositionSolution<T>&
  decompositionSolution<T>::copy(const decompositionSolution<T>& other) {
    linearEquationSystemSolutionMethod<T>::copy(other);
    decomposed=other.decomposed;
    dcmpMat=other.dcmpMat;

    return (*this);
  }

  //--- QR solution functor ---
  //---------------------------

  // returns the current parameters
  template<class T>
  const typename qrSolution<T>::parameters&
  qrSolution<T>::getParameters() const {
    const parameters* params =
      dynamic_cast<const parameters*>(&functor::getParameters());
    if (isNull(params)) {
      throw functor::invalidParametersException(getTypeName());
    }
    return *params;
  }

  // onPlace version of apply
  template<class T>
  double qrSolution<T>::apply(vector<T>& b) {
    vector<T> tmpVec;
    double tmpRes;
    tmpRes=apply(b,tmpVec);
    b=tmpVec;
    return tmpRes;
  }

  // onCopy version of apply
  // from: Engeln-Mllges, Reutter
  //       Formelsammlung zur Numerischen Mathematik mit C-Programmen
  //       BI-Wissenschaftsverlag, 1990
  //       p. 492
  template<class T>
  double qrSolution<T>::apply(const vector<T>& b,vector<T>& x) {
    const parameters& tmpParam = getParameters();
    const matrix<T>& A = tmpParam.systemMatrix;
    const int& n = A.columns();
    const int& m = A.rows();
    vector<T> c(b);
    double radiant,alpha,factor,sum;
    int i,k,j;

    //        int si=b.size();
    if ((m < n)||(m != b.size())) {
      linearAlgebraFunctor::incompatibleDimensionsException tmpExc;
      throw tmpExc;
    }

    if(!decomposed) {
      // decompose A, compose R=HA, compose c=Hb (H: transposed Q-matrix)
      dcmpMat.castFrom(A);
      dcmpVec.resize(n);
      helpVec.resize(n);
      for(i = 0; i < n; i++) {
        for(radiant = dcmpMat[i][i]*dcmpMat[i][i], k = i + 1; k < m; k++) {
          radiant += dcmpMat[k][i] * dcmpMat[k][i];
        }
        if(radiant < (4*std::numeric_limits<double>::epsilon())) {
          linearAlgebraFunctor::singularMatrixException tmpExc;
          throw tmpExc;
        }
        if(dcmpMat[i][i] > 0) {
          alpha = sqrt(radiant);
        }
        else {
          alpha =- sqrt(radiant);
        }
        helpVec[i] = 1.0 / (radiant + alpha * dcmpMat[i][i]);
        dcmpMat[i][i] += static_cast<T>(alpha);

        dcmpVec[i] =- alpha;
        for(k = i + 1; k < n; k++) {
          for(factor = .0 ,j = i;j < m;j++) {
            factor += dcmpMat[j][k] * dcmpMat[j][i];
          }
          factor *= helpVec[i];
          for(j = i; j < m; j++) {
            dcmpMat[j][k] -= static_cast<T>(factor * dcmpMat[j][i]);
          }
        }
        for(factor = .0 ,j = i;j < m;j++) {
          factor += c[j] * dcmpMat[j][i];
        }
        factor *= helpVec[i];
        for(j = i; j < m; j++) {
          c[j] -= static_cast<T>(factor * dcmpMat[j][i]);
        }
      }
      decomposed = true;
    }
    else { // compose c=Hb
      for(i = 0; i < n; i++) {
        for(factor = .0, j = i; j < m; j++) {
          factor += c[j] * dcmpMat[j][i];
        }
        factor *= helpVec[i];
        for(j = i; j < m; j++) {
          c[j] -= static_cast<T>(factor * dcmpMat[j][i]);
        }
      }
    }

    // solve Rx=c
    x.resize(n);
    for(i = n-1; i >= 0; i--) {
      for(sum = .0, k = i + 1; k < n; k++) {
        sum += dcmpMat[i][k] * x[k];
      }
      x[i] = static_cast<T>((c[i] - sum) / dcmpVec[i]);
    }

    if(tmpParam.computeResiduum) { // calculate residuum
      for(sum = .0, i = n; i < m; i++) {
        sum += c[i]*c[i];
      }
      return sqrt(sum);
    }

    return -1;
  }

  // copy data of "other" functor.
  template<class T>
  qrSolution<T>& qrSolution<T>::copy(const qrSolution<T>& other) {
    decompositionSolution<T>::copy(other);
    helpVec=other.helpVec;
    dcmpVec=other.dcmpVec;

    return (*this);
  }

  //--- LU solution functor ---
  //---------------------------

  // onPlace version of apply
  // see for example: Press, Vetterling, Teukolsky, Flannery
  //                  Numerical Recipes in C, 2nd edition
  //                  Cambridge University Press, 1992
  //                  p. 47
  template<class T>
  bool luSolution<T>::apply(vector<T>& b) {
    const parameters& tmpParam = getParameters();
    const int& n=tmpParam.systemMatrix.rows();
    if((n != tmpParam.systemMatrix.columns()) || (n!=b.size())) {
      setStatusString("Incompatible dimensions");
      return false;
    }
    int i,ii=-1,ip,j;
    T sum;

    // lu decomposition
    if(!decomposed) {
      luDecomposition< T > decmp;
      decmp.apply(tmpParam.systemMatrix,dcmpMat,dcmpVec);
      decomposed = true;
    }

    // forward substitution and backsubstitution
    for (i=0;i<n;i++) {
      ip=dcmpVec[i];
      sum=b[ip];
      b[ip]=b[i];
      if (ii+1) {
        for (j=ii;j<=i-1;j++) {
          sum -= dcmpMat[i][j]*b[j];
        }
      }
      else if (sum) {
        ii=i;
      }
      b[i]=sum;
    }
    for (i=n-1;i>=0;i--) {
      sum=b[i];
      for (j=i+1;j<n;j++) {
        sum -= dcmpMat[i][j]*b[j];
      }
      b[i]=sum/dcmpMat[i][i];
    }

    return true;
  }

  // onCopy version of apply
  template<class T>
  bool luSolution<T>::apply(const vector<T>& b,vector<T>& x) {
    x.copy(b);
    return apply(x);
  }

  // solve many equations at the same time!
  template<class T>
  bool luSolution<T>::apply(matrix<T>& X) {
    bool result=false;
    X.transpose();
    for (int i=0;i<X.rows();i++) {
      result=apply(X.getRow(i));
    }
    X.transpose();
    return result;
  }

  // solve many equations at the same time!
  template<class T>
  bool luSolution<T>::apply(const matrix<T>& B,matrix<T>& X) {
    X.copy(B);
    return apply(X);
  }

  // copy data of "other" functor.
  template<class T>
  luSolution<T>& luSolution<T>::copy(const luSolution<T>& other) {
    decompositionSolution<T>::copy(other);
    dcmpVec=other.dcmpVec;

    return (*this);
  }
}

