/*----------------------------------------------------------------
 * project ....: LTI Digital Image/Signal Processing Library
 * file .......: ltiHmmViterbiPathSearch.h
 * authors ....: Benjamin Winkler
 * organization: LTI, RWTH Aachen
 * creation ...: 12.9.2001
 * revisions ..: $Id: ltiHmmViterbiPathSearch.h,v 1.10 2003/02/17 07:16:42 alvarado Exp $
 */

#ifndef _LTI_HMM_VITERBI_PATH_SEARCH_H_
#define _LTI_HMM_VITERBI_PATH_SEARCH_H_

#include "ltiSequence.h"
#include "ltiVector.h"
#include "ltiMatrix.h"
#include "ltiHiddenMarkovModel.h"

namespace lti {

  /**
   * Finds and evaluates the best path through a Hidden Markov Model (HMM)
   * for a given observation sequence. The best path is found with the
   * Viterbi algorithm.
   *
   * See: L. Rabiner and B.-H. Juang, "Fundamentals of speech recognition"
   * Prentice Hall, 1993.
   */
  class hmmViterbiPathSearch {
  public:

    /**
     * type of an observation sequence
     */
    typedef sequence<dvector> observationSequence;

    /**
     * default constructor
     */
    hmmViterbiPathSearch();

    /**
     * copy constructor
     * @param other the object to be copied
     */
    hmmViterbiPathSearch(const hmmViterbiPathSearch& other);

    /**
     * destructor
     */
    virtual ~hmmViterbiPathSearch();

    /**
     * returns the name of this type ("hmmViterbiPathSearch")
     */
    virtual const char* getTypeName() const;

    /**
     * copy data of "other" functor.
     * @param other the functor to be copied
     * @return a reference to this functor object
     */
    hmmViterbiPathSearch& copy(const hmmViterbiPathSearch& other);



    /**
     * Performs the Viterbi path search on a given (observed) data sequence and
     * returns its score.
     * @param model the Hidden Markov Model.
     * @param srcData contains the sequence of observed vectors.
     * @result score of the best path through the model for the given sequence.
     */
    double apply(const hiddenMarkovModel &model, const observationSequence &srcData);

    /**
     * Calculates score along a given path for a given (observed) data sequence.
     * @param model the Hidden Markov Model.
     * @param srcData contains the sequence of observed vectors.
     * @param path defines the applicable state sequence.
     * @result score of the given path.
     */
    double evaluatePath(const hiddenMarkovModel &model, const observationSequence &srcData, const ivector &path);

    /**
     * Calculates score along a given path for a given (observed) data sequence,
     * and additionally selects the specified density centers of the mixture densities
     * in the visited path.
     * @param model the hidden markov model
     * @param srcData contains the data sequence.
     * @param path defines the applicable state sequence.
     * @param densityPath contains the applicable mixture density center selection.
     * @result score of the given path.
     */
    double evaluatePath(const hiddenMarkovModel &model, const observationSequence &srcData,
                       const ivector &path, const ivector &densityPath);

    /**
     * Find out lowest possible score for the given model and a sequence of the specified length.
     * @param model the hidden markov model
     * @param pathLength length of the sequence
     * @return minimum score
     */
    double minimumPathScore(const hiddenMarkovModel &model, int pathLength) const;


    /**
     * find out lowest possible score of the given model for a given path.
     * @param model the hidden markov model
     * @param stateSelection the state selection along the given path
     * @param densitySelection the density selection along the given path. if this is empty, the best densities are chosen.
     * @return the minimum score of the given path.
     */
    double minimumPathScore(const hiddenMarkovModel &model, const ivector &stateSelection, const ivector &densitySelection = ivector()) const;

    /** @name State selection logging
     *  These methods can be used to enable/disable logging and access the selection of
     *  states during a call of apply().
     *  Default is "disabled", for efficiency reasons. Enable only if necessary.
     */
      //@{

        /**
         * Turns logging of state selection on and off.
         * This enables to review the state matrix and the path leading to the minimal score.
         *
         * By default, this is turned off for efficiency reasons (memory).
         */
        void logStateSelection(bool enable = true);

        /**
         * @return matrix of the best states leading to the minimal score.
         * If logging of state selection is turned off or a given path was evaluated, the matrix will be empty.
         */
        const matrix<int>& getBackpointerMatrix() const;

        /**
         * @return chosen path of the last run leading to the minimal score
         * If logging of state selection is turned off or a given path was evaluated, the vector will be empty.
         */
        const vector<int>& getStatePath();

      //@}


    /** @name Density selection logging
     *  These methods can be used to enable/disable logging and access the selection of
     *  densities during a call of apply().
     *  Default is "disabled", for efficiency reasons. Enable only if necessary.
     */
      //@{

        /**
         * Turns logging of density selection matrix on and off.
         * This enables to review the densities selected during evaluation of a sequence.
         *
         * By default, this is turned off for efficiency reasons (memory).
         */
        void logDensitySelection(bool enable = true);

        /**
         * @return matrix of the densities used in the last pass.
         * If logging of density selection is turned off or a given path was evaluated, the matrix will be empty.
         */
        const matrix<int>& getDensitySelectionMatrix() const;

        /**
         * @return densities selected for the chosen path.
         * If logging of density selection is turned off, the vector will be empty.
         */
        const vector<int>& getDensitySelectionPath();

      //@}


    /** @name Emission score logging
     *  These methods can be used to enable/disable logging and access the emission
     *  scores during a call of apply().
     *  Default is "disabled", for efficiency reasons. Enable only if necessary.
     */
      //@{

        /**
         * Turns logging of emission score matrix on and off.
         * This enables to analyse the emission scores generated for each step and state
         *
         * By default, this is turned off for efficiency reasons (runtime and memory).
         */
        void logEmissionScore(bool enable = true);

        /**
         * @return matrix of the emission scores generated in the last pass.
         * If logging of emission scores is turned off or a given path was evaluated, the matrix will be empty.
         */
        const matrix<double>& getEmissionScoreMatrix() const;

        /**
         * @return emission scores generated for the chosen path.
         * If logging of emission score is turned off, the vector will be empty.
         */
        const vector<double>& getEmissionScorePath();

      //@}

    /**
     * @name Routines for online classification
     */
      //@{

        /**
         * perform initial transition and add emission score
         * @return last entry of result vector
         */
        double initialStep(const hiddenMarkovModel &model, const dvector &frame, dvector &result) const;

        /**
         * perform subsequent transition and add emission score
         * @return last entry of result vector
         */
        double subsequentStep(const hiddenMarkovModel &model, const dvector &frame, dvector &result) const;


      //@}


  private:


    /**
     * clear logged data
     */
    void reset();


    /**
     * add initial transition vector to result
     * adding is done to allow for later implementation of observation chain recognition
     */
    void initialTransition(const hiddenMarkovModel &model, dvector &result) const;

    /**
     * calculate best transitions
     */
    void subsequentTransition(const hiddenMarkovModel &model, dvector &result, ivector &predecessors) const;

    /**
     * calculate emission score for given frame and add it to the result vector
     */
    void addEmissionScore(const hiddenMarkovModel &model, const dvector &frame,
                          dvector &result, ivector &densitySelection, dvector &scoreSelection) const;

    /**
     * perform initial transition and add emission score
     * @return last entry of result vector
     */
    double initialStep(const hiddenMarkovModel &model, const dvector &frame, dvector &result,
	                    ivector &densitySelection, dvector &scoreSelection) const;

    /**
     * perform subsequent transition and add emission score
     * @return last entry of result vector
     */
    double subsequentStep(const hiddenMarkovModel &model, const dvector &frame, dvector &result,
	                       ivector &stateSelection, ivector &densitySelection, dvector &scoreSelection) const;

    // score functions

    /**
     * calculate score for given frame and density
     */
    double getDensityScore(const hiddenMarkovModel &model, const dvector &frame, const hiddenMarkovModel::singleDensity &dens) const;

    /**
     * calculate score for given frame and state, return best density
     */
    double getStateScoreAndDensity(const hiddenMarkovModel &model, const dvector &frame, const hiddenMarkovModel::mixtureDensity &state, int &dens) const;

    /**
     * gaussian Score (i.e. -ln(N(mean, variance)), where N is the normal distribution)
     */
    double gaussScore(const dvector &frame, const dvector &mean,
                     const dvector &variance, const dvector &featureWeights) const;

    /**
     * laplace Score (i.e. -ln(L(mean, variance)), where L is the laplace distribution)
     */
    double laplaceScore(const dvector &frame, const dvector &mean,
                       const dvector &variance, const dvector &featureWeights) const;

    /**
     * defines, whether state selection is logged
     */
    bool registerStateSelection;

    /**
     * backtrace matrix will be saved here
     */
    matrix<int> stateMatrix;

    /**
     * best path will be generated by getStatePath and then saved here
     */
    vector<int> statePath;


    /**
     * defines, whether density selection is logged
     */
    bool registerDensitySelection;

    /**
     * density selection matrix will be saved here
     */
    matrix<int> densityMatrix;

    /**
     * density selection path will be generated by getDensityPath and then saved here
     */
    vector<int> densityPath;


    /**
     * defines, whether emission scores are logged
     */
    bool registerEmissionScore;

    /**
     * emission score matrix will be saved here
     */
    matrix<double> emissionScoreMatrix;

    /**
     * emission score path will be generated by getEmissionScorePath and then saved here
     */
    vector<double> emissionScorePath;


//    friend class hmmTrainer;

  private:

    int stateCount;

  };

}

#endif

