/*
 * Copyright (C) 1998, 1999, 2000, 2001
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 *
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.
 */


/*----------------------------------------------------------------
 * project ....: LTI Digital Image/Signal Processing Library
 * file .......: ltiSemaphore.cpp
 * authors ....: Thomas Rusert
 * organization: LTI, RWTH Aachen
 * creation ...: 04.11.99
 * revisions ..: $Id: ltiSemaphore.cpp,v 1.8 2003/10/30 10:03:41 alvarado Exp $
 */

#include "ltiSemaphore.h"
#include "ltiAssert.h"
#include "ltiException.h"
#include <errno.h>

namespace lti {

  // --------------------------------------------------------
  //                        UNIX/LINUX
  // --------------------------------------------------------

  // returns the name of this type
  const char* semaphore::getTypeName() const {
    return "semaphore";
  }

# ifndef _LTI_WIN32

  // default constructor
  semaphore::semaphore(const int initialValue) :
    initValue(initialValue) {
    assert( initValue<SEM_VALUE_MAX );
    sem_init(&theSemaphore,0,initValue);
  }

  // destructor
  semaphore::~semaphore() {
    destroy();
  }

  // wait on semaphore, i.e. decrease the value or wait if counter <= 0
  bool semaphore::wait() {
    return (sem_wait(&theSemaphore) == 0);
  }

  // try to wait on semaphore, but do not block
  bool semaphore::tryWait() {
    return (sem_trywait(&theSemaphore)==0);
  }

  // post semaphore, i.e. increase the value
  bool semaphore::post() {
    return (sem_post(&theSemaphore) == 0);
  }

  // get current value
  const int semaphore::getValue() {
    int value;
    sem_getvalue(&theSemaphore,&value);
    return value;
  }

  // reset value to initialValue
  void semaphore::reset() {
    destroy();
    sem_init(&theSemaphore,0,initValue);
  }

  // destroy semaphore
  void semaphore::destroy() {
    while(sem_destroy(&theSemaphore)!=0) {
      assert( errno==EBUSY );
      while(getValue()<=0) {
        post();
      }
    }
    sem_destroy(&theSemaphore);
  }

# else
  // --------------------------------------------------------
  //                        WINDOWS
  // --------------------------------------------------------

  // default constructor
  semaphore::semaphore(const int initialValue)
    : initValue(initialValue),counter(initialValue) {
    theSemaphore =
       CreateSemaphore(NULL,initialValue,SEM_VALUE_MAX,NULL);
    if (theSemaphore == NULL) {
      unsigned int error;
      error = GetLastError();
      throw exception("Semaphore cannot be created!");
    }
  }

  // destructor
  semaphore::~semaphore() {
    destroy();
  }

  // wait until lock for semaphore becomes available
  bool semaphore::wait() {
    counter--;
    return (WaitForSingleObject(theSemaphore,INFINITE) != WAIT_FAILED);
  }

  // try to lock semaphore, but do not block
  bool semaphore::tryWait() {
    if (WaitForSingleObject(theSemaphore,0) == WAIT_OBJECT_0) {
      counter--;
      return true;
    } else {
      return false;
    };
  }

  // unlock semaphore
  bool semaphore::post() {
    // TODO: This value might be invalid:
    counter++;
    return (ReleaseSemaphore(theSemaphore,1,NULL) != 0);
  }

  // destroy semaphore
  void semaphore::destroy() {

    while(!tryWait() && (counter < initValue) ) {
      post();
    }

    post();

    CloseHandle(theSemaphore);
  }
  // get current value
  const int semaphore::getValue() {
    return counter;
  }

  // reset value to initialValue
  void semaphore::reset() {
    destroy();
    theSemaphore =
       CreateSemaphore(NULL,initValue,SEM_VALUE_MAX,NULL);

  }

# endif
}

